/*
 * PANDA -- a simple transaction monitor
 * Copyright (C) 2000-2009 Ogochan & JMA (Japan Medical Association).
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/*
#define	CONSOLE_DEBUG
#define	TRACE
#define	DEBUG
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>
#include	<setjmp.h>
#include	"libs.h"
#include	"lex.h"
#include	"compile.h"
#include	"hakama.h"
#include	"debug.h"

#define	T_CREATE		(T_YYBASE +1)
#define	T_INSERT		(T_YYBASE +2)
#define	T_DELETE		(T_YYBASE +3)
#define	T_UPDATE		(T_YYBASE +4)
#define	T_SEARCH		(T_YYBASE +5)
#define	T_SELECT		(T_YYBASE +6)
#define	T_SET			(T_YYBASE +7)
#define	T_DROP			(T_YYBASE +8)
#define	T_DECLARE		(T_YYBASE +9)
#define	T_CLOSE			(T_YYBASE +10)
#define	T_FETCH			(T_YYBASE +11)

#define	T_START			(T_YYBASE +20)
#define	T_COMMIT		(T_YYBASE +21)
#define	T_ABORT			(T_YYBASE +22)

#define	T_TABLE			(T_YYBASE +30)
#define	T_INDEX			(T_YYBASE +31)
#define	T_DICTIONARY	(T_YYBASE +32)
#define	T_LOCALE		(T_YYBASE +33)
#define	T_TRIGGER		(T_YYBASE +34)
#define	T_CURSOR		(T_YYBASE +35)
#define	T_LIMIT			(T_YYBASE +36)
#define	T_OFFSET		(T_YYBASE +37)

#define	T_ON			(T_YYBASE +40)
#define	T_USING			(T_YYBASE +41)
#define	T_OPTION		(T_YYBASE +42)
#define	T_INTO			(T_YYBASE +43)
#define	T_VALUES		(T_YYBASE +44)
#define	T_FROM			(T_YYBASE +45)
#define	T_KEY			(T_YYBASE +46)
#define	T_WHERE			(T_YYBASE +47)
#define	T_FOR			(T_YYBASE +48)
#define	T_AS			(T_YYBASE +49)
#define	T_ORDER			(T_YYBASE +50)
#define	T_BY			(T_YYBASE +51)
#define	T_FORWARD		(T_YYBASE +52)
#define	T_BACKWARD		(T_YYBASE +53)
#define	T_RELATIVE		(T_YYBASE +54)
#define	T_IN			(T_YYBASE +55)
#define	T_DESC			(T_YYBASE +56)
#define	T_ASC			(T_YYBASE +57)
#define	T_ALL			(T_YYBASE +58)
#define	T_DISTINCT		(T_YYBASE +59)

#define	T_NULL			(T_YYBASE +60)

#define	T_AND			(T_YYBASE +70)
#define	T_OR			(T_YYBASE +71)
#define	T_NOT			(T_YYBASE +72)

#define	T_CROSS			(T_YYBASE +80)
#define	T_INNER			(T_YYBASE +81)
#define	T_OUTER			(T_YYBASE +82)
#define	T_JOIN			(T_YYBASE +83)
#define	T_LEFT			(T_YYBASE +84)
#define	T_RIGHT			(T_YYBASE +85)
#define	T_FULL			(T_YYBASE +86)

#define	T_LIKE			(T_YYBASE +90)
#define	T_BETWEEN		(T_YYBASE +91)

#define	T_CHAR			(T_YYBASE +101)
#define	T_TEXT			(T_YYBASE +102)
#define	T_INT			(T_YYBASE +103)
#define	T_BOOL			(T_YYBASE +106)
#define	T_BYTE			(T_YYBASE +107)
#define	T_NUMBER		(T_YYBASE +108)
#define	T_VARCHAR		(T_YYBASE +109)
#define	T_DBCODE		(T_YYBASE +110)
#define	T_OBJECT		(T_YYBASE +111)
#define	T_FLOAT			(T_YYBASE +112)
#define	T_TIMESTAMP		(T_YYBASE +113)
#define	T_DATE			(T_YYBASE +114)
#define	T_TIME			(T_YYBASE +115)
#define	T_ENUM			(T_YYBASE +116)
#define	T_BINARY		(T_YYBASE +119)
#define	T_UNIQ			(T_YYBASE +120)
#define	T_PRIMARY		(T_YYBASE +121)

static jmp_buf compiler_env;


#ifdef	Error
#undef	Error
#endif
#define Error(...)	ErrorBasicParser(in, __VA_ARGS__)
#define	CHECK_TERMINATOR					\
{											\
	if		(  ComToken  !=  T_EOF  ) {		\
		if 		(  ComToken  !=  ';'  ) {	\
			Error("missing ;(semicolon).");	\
		}									\
	}										\
}

static	void	ParExpr(HakamaRecord *expr, CURFILE *in, char *tablename);
static	void	ParItemDefines(CURFILE *in, ItemStruct *upper);

extern	void
ErrorBasicParser(
	CURFILE	*in,
	char	*format,
	...)
{
#ifdef DEBUG
	va_list	va;

	in->fError = TRUE;
	fprintf(stderr, "Hakama ODBC: %d:", in->cLine);
	va_start(va,format);
	vfprintf(stderr,format,va);
	va_end(va);
	fprintf(stderr, "\n");
#endif
	longjmp(compiler_env, 1);
}

static	SqlInfo	*
NewSqlInfo(
	int		type)
{
	SqlInfo	*ret;

	ret = New(SqlInfo);
	ret->type = type;
	ret->tname = NULL;
	ret->rname = NULL;
	ret->pname = NULL;
	ret->rec = NULL;
	ret->fields = NULL;
	ret->record = NULL;
	ret->limit = 0;
	ret->offset = 0;

	return	(ret);
}

static	char	*
ParName(
	CURFILE			*in)
{
	char	name[SIZE_LONGNAME+1];
	char	*ret
		,	*p;

ENTER_FUNC;
	*name = 0;
	p = name;
	while	(	(  ComToken  ==  T_SYMBOL  )
			||	(  ComToken  ==  T_STRING  ) ) {
		strcpy(p,ComSymbol);
		p += strlen(p);
		if		(  GetSymbol  ==  '.'  ) {
			*p ++ = '.';
			GetName;
		} else
			break;
	}
	if		(  *name  !=  0  ) {
		ret = StrDup(name);
	} else {
		ret = NULL;
	}
LEAVE_FUNC;
	return	(ret);
}

static	char	*
ItemName(
	char	*longname)
{
	char	*name
		,	*p;

	if		(  ( p = strchr(longname,'.') )  !=  NULL  ) {
		name = StrDup(p+1);
	} else {
		name = StrDup(longname);
	}
	return	(name);
}

static	Fields	*
ParItemsName(
	     CURFILE			*in,
	     int next_token)
{
	char	*name;
	Fields	*ret;

ENTER_FUNC;
	ret = New(Fields);
	ret->n = 0;
	ret->name = NULL;
	if		(  ComToken  ==  '*'  ) {
		GetSymbol;
		ret->n = -1;
	} else {
		while	(	(  ComToken  ==  T_SYMBOL  )
				||	(  ComToken  ==  T_STRING  ) ) {
			name = ParName(in);
			ret->name = (char **)xrealloc(ret->name, sizeof(char *) * ( ret->n + 1));
			ret->name[ret->n] = ItemName(name);
			xfree(name);
			dbgprintf("name  = [%s]", ret->name[ret->n]);
			ret->n ++;
			if		(  ComToken  ==  ','  ) {
				GetName;
				/*	OK	*/
			} else
			if		(	(  ComToken  ==  next_token  )
					||	(  ComToken  ==  ')'     ) ) {
			} else {
				Error(", missing");
			}
		}
	}
	if		(	(  !in->fError      )
			&&	(	(  ComToken  ==  next_token  )
				||	(  ComToken  ==  ')'     ) ) ) {
		/*	OK	*/
	} else {
		dbgprintf("token = %d [%c]",ComToken,ComToken);
		if (ret) {
		  DestroyFields(ret);
		}
		Error("syntax error");
	}
LEAVE_FUNC;
	return	(ret);
}

Fields *
ParOrderBy(
	   CURFILE			*in)
{
	char	*name;
	Fields	*ret;

ENTER_FUNC;
	ret = New(Fields);
	ret->n = 0;
	ret->name = NULL;
	while	(	(  ComToken  ==  T_SYMBOL  )
			||	(  ComToken  ==  T_STRING  ) ) {
	  name = ParName(in);
	  ret->name = (char **)xrealloc(ret->name, sizeof(char *) * ( ret->n + 1));
	  ret->name[ret->n] = ItemName(name);
	  xfree(name);
	  dbgprintf("name  = [%s]", ret->name[ret->n]);
	  ret->n ++;
	  if		(  ComToken  ==  ','  ) {
	    GetName;
	    /*	OK	*/
	  } else if ((ComToken  ==  T_DESC  )
		     || (  ComToken  ==  T_ASC  )) {
	    GetName;
	    if		(  ComToken  ==  ','  ) {
	      GetName;
	    }
	  } else if ((  ComToken  ==  T_LIMIT  )
		     || (  ComToken  ==  T_EOF  )				
		     ||	(  ComToken  ==  ')'     )) {
	  } else {
	    Error(", missing");
	  }
	}
	if		(	(  !in->fError      )
				&&	(	(  ComToken  ==  T_DESC  )
						|| (  ComToken  ==  T_ASC  )
						|| (  ComToken  ==  T_LIMIT  )
						|| (  ComToken  ==  T_EOF  )						
						||	(  ComToken  ==  ')'     ) ) ) {
	  /*	OK	*/
	} else {
	  dbgprintf("token = %d [%c]",ComToken,ComToken);
	  Error("syntax error");
	}
LEAVE_FUNC;
	return	(ret);
}

extern	char	*
ParValue(
	CURFILE		*in)
{
	char	buff[SIZE_LONGNAME+1];
	char	*value = NULL;

ENTER_FUNC;
	switch	(ComToken) {
	case	T_ICONST:
	  value = StrDup(ComSymbol);
	  GetSymbol;
	  break;
	case	T_SCONST:
	  sprintf(buff,"%s",ComSymbol);
	  value = StrDup(buff);
	  GetSymbol;
	  break;
	case T_NULL:
	  buff[0] = '\x01';
	  buff[1] = '\x00';
	  //sprintf(buff,"NULL");
	  value = StrDup(buff);
	  dbgmsg("null");
	  GetSymbol;
	  break;
	default:
	  dbgprintf("? token = %d\n",ComToken-T_YYBASE);
	  value = NULL;
	  GetSymbol;
	  break;
	}
	dbgprintf("token = %s\n", value);	
LEAVE_FUNC;
	return	(value);
}

static	Fields	*
ParItemsValue(
	CURFILE			*in)
{
  char	*value;
  Fields	*ret;
  
  ENTER_FUNC;
  ret = New(Fields);
  ret->n = 0;
  ret->name = NULL;
  while ((ComToken == T_SYMBOL  )
	 || (ComToken == T_STRING)
	 || (ComToken == T_NULL)			
	 || (ComToken == T_ICONST)
	 || (ComToken == T_SCONST))
  {
    value = ParValue(in);
    dbgprintf("value  = [%s]\n", value);		
    ret->name = (char **)xrealloc(ret->name, sizeof(char *) * ( ret->n + 1));
    ret->name[ret->n] = value;
    dbgprintf("ret value  = [%s]\n", ret->name[ret->n]);
    ret->n ++;
    if (ComToken == ',') {
      GetSymbol;
      /*	OK	*/
    } else
      if (ComToken == ')') {
	/*	OK	*/
      } else {
	Error(", missing");
      }
  }
  if ((!in->fError)
      &&((ComToken == ')')))
  {
      /*	OK	*/
  } else {
    dbgprintf("token = %d [%c]",ComToken,ComToken);
    Error("syntax error");
  }
LEAVE_FUNC;
	return	(ret);
}


static	void
ParRelationExpr(
	HakamaRecord	*expr,
	CURFILE			*in,
	char			*tablename)
{
	char	*name
		,	*p;
	char	*value;
	char	buff[SIZE_LONGNAME+1];
ENTER_FUNC;
	if		(  ComToken  ==  '('  ) {
		GetSymbol;
		ParExpr(expr,in,tablename);
		if		(  ComToken  ==  ')'  ) {
			GetSymbol;
		} else {
			Error("factor invalid (");
		}
	} else {
		p = ParName(in);
		/*	relation operation	*/
		GetSymbol;
		if		(  p  !=  NULL  ) {
			name = ItemName(p);
			xfree(p);
			value = ParValue(in);
			sprintf(buff,"%s.%s",tablename,name);
			HakamaEnterValue(expr,StrDup(buff),value);
			dbgprintf("%s = %s",name,value);
			xfree(name);
		}
	}
LEAVE_FUNC;
}

static	void
ParNotExpr(
	HakamaRecord	*expr,
	CURFILE			*in,
	char			*tablename)
{
	Bool	fNot;

ENTER_FUNC;
	fNot = (  ComToken  ==  T_NOT  ) ? TRUE : FALSE;
	ParRelationExpr(expr,in,tablename);
LEAVE_FUNC;
}

static	void
ParAndExpr(
	HakamaRecord	*expr,
	CURFILE			*in,
	char			*tablename)
{
ENTER_FUNC;
	ParNotExpr(expr,in,tablename);
	while	(  ComToken  ==  T_AND  ) {
		GetSymbol;
		ParNotExpr(expr,in,tablename);
	}
LEAVE_FUNC;
}

static	void
ParExpr(
	HakamaRecord	*expr,
	CURFILE			*in,
	char			*tablename)
{
ENTER_FUNC;
	ParAndExpr(expr,in,tablename);
	while	(  ComToken  ==  T_OR  ) {
		GetSymbol;
		ParAndExpr(expr,in,tablename);
	}
LEAVE_FUNC;
}

static	void
ParSequence(
	HakamaRecord	*expr,
	CURFILE			*in,
	char			*tablename)
{
	Bool	fExit;

ENTER_FUNC;
	fExit = FALSE;
	do {
		ParExpr(expr,in,tablename);
		switch	(ComToken) {
		  case	',':
		  case	')':
			if		(  ComToken  ==  ')'  ) {
				fExit = TRUE;
			} else {
				GetSymbol;
			}
			break;
		  default:
			fExit = TRUE;
			break;
		}
	}	while	(  !fExit  );
LEAVE_FUNC;
}

static	HakamaRecord	*
ParPredicate(
	CURFILE		*in,
	char		*tablename)
{
	HakamaRecord	*expr;

ENTER_FUNC;
	GetSymbol;
	expr = HakamaNewRecord();
	ParSequence(expr,in,tablename);
LEAVE_FUNC;
	return	(expr);
}


static	SqlInfo	*
ParSelect(
	CURFILE			*in)
{
	SqlInfo		*ret;
	char		*tablename;
	Fields		*fields;
	Fields		*order = NULL;
	
	HakamaRecord	*predicate = NULL;
	char field_name[SIZE_LONGNAME + 1];
	char		*p
		,		*tname
		,		*rname
		,		*pname
		,		*name;
	int			limit
		,		offset
		,		i;

ENTER_FUNC;
	GetName;
	fields = ParItemsName(in, T_FROM);
	tablename = NULL;
	if		(  ComToken  ==  T_FROM  ) {
		GetName;
		if		(	(  ComToken  !=  T_SYMBOL  )
				&&	(  ComToken  !=  T_STRING  ) ) {
			Error("table name invalid");
		}
		while	(	(  ComToken  ==  T_SYMBOL  )
				||	(  ComToken  ==  T_STRING  ) ) {
			name = ParName(in);
			if		(  tablename  ==  NULL  ) {
				if		(  ( p = strchr(name,'.') )  !=  NULL  ) {
					tablename = StrDup(p+1);
				} else {
					tablename = StrDup(name);
				}
			}
			xfree(name);
			if		(	(  ComToken  !=  T_SYMBOL  )
					&&	(  ComToken  !=  T_STRING  ) ) {
				if		(  ComToken  !=  ','  )	break;
				GetName;
			}
		}
		dbgprintf("tablename = [%s]",tablename);
	} else {
	  Error("FROM missing");
	}
	dbgmsg("*");
	tname = StrDup(tablename);
	if		(  ( p = strchr(tablename,'$') )  !=  NULL  ) {
		*p = 0;
		rname = StrDup(tablename);
		pname = StrDup(p+1);
	} else {
		rname = StrDup(tablename);
		pname = StrDup("");
	}
	xfree(tablename);
	tablename = NULL;
	predicate = NULL;
	limit = -1;
	offset = 0;
	while	(	(  ComToken  ==  T_WHERE   )
			||	(  ComToken  ==  T_LIMIT   )
			||	(  ComToken  ==  T_ORDER   )			
			||	(  ComToken  ==  T_OFFSET  ) ) {
		switch	(ComToken)	{
		  case	T_WHERE:
			predicate = ParPredicate(in,rname);
			break;
		  case	T_LIMIT:
			if		(  GetSymbol  ==  T_ICONST  ) {
				limit = atoi(ComSymbol);
				GetSymbol;
			} else {
			  Error("invalid limit");
			}
			break;
		  case	T_OFFSET:
			if		(  GetSymbol  ==  T_ICONST  ) {
				offset = atoi(ComSymbol);
				GetSymbol;
			} else {
			  Error("invalid offset");
			}
			break;
		case	T_ORDER:
			dbgmsg("order");
			if		(  GetSymbol  ==  T_BY  ) {
			  dbgmsg("by");
			  GetName;
			  order = ParOrderBy(in);
			  for (i = 0; i < order->n; ++i) {
			    dbgprintf("order [%s]", order->name[i]);
			  }
			  switch (ComToken) {
			  case T_DESC:
			  case T_ASC:
			    GetSymbol;
			    break;
			  default:
			    break;
			  }
			  DestroyFields(order);
			  order = NULL;
			} else {
			  Error("invalid offset");
			}
			break;
		}
	}
	
	if (fields && fields->n >= 0) {
	  for (i = 0; i < fields->n; ++i) {
	    sprintf(field_name, "%s$%s.%s", rname, pname, fields->name[i]);
	    xfree(fields->name[i]);
	    fields->name[i] = StrDup(field_name);
	  }
	}
	
	dbgprintf("* ComToken = %d, error = %d", ComToken, in->fError);
	CHECK_TERMINATOR;
	if		(  !in->fError  ) {
		ret = NewSqlInfo(SQL_TYPE_SELECT);
		ret->tname = tname;
		ret->rname = rname;
		ret->pname = pname;
		ret->rec = predicate;
		ret->fields = fields;
		ret->limit = limit;
		ret->offset = offset;
	} else {
		ret = NULL;
		dbgmsg("error");
	}
LEAVE_FUNC;
	return	(ret);
}


/*
  UPDATE TABLE_NAME SET A=2,B=C,D=HOGE WHERE A=B AND C=1
*/
static	SqlInfo	*
ParUpdate(
	CURFILE			*in)
{
  SqlInfo	*ret;
  char		*tablename;
  HakamaRecord	*rec;
  HakamaRecord	*predicate;
  char		*p
    ,		*tname
    ,		*rname
    ,		*pname
    ,		*name;
  
ENTER_FUNC;
  GetName;
  if ((ComToken !=  T_SYMBOL)
      && (ComToken !=  T_STRING)) {
      Error("table name invalid");
  } else {
    name = ParName(in);
    if (tablename  ==  NULL) {
      if ((p = strchr(name,'.'))  !=  NULL  ) {
	tablename = StrDup(p + 1);
      } else {
	tablename = StrDup(name);
      }
    }
    xfree(name);
  }
  dbgprintf("tablename = [%s]\n", tablename);
  tname = StrDup(tablename);
  if ((p = strchr(tablename,'$')) !=  NULL) {
    *p = 0;
    rname = StrDup(tablename);
    pname = StrDup(p+1);
  } else {
    rname = StrDup(tablename);
    pname = StrDup("");
  }
  xfree(tablename);
  tablename = NULL;
  
  dbgmsg("*");
  if (ComToken != T_SET) {
    Error("SET missing");
  }
  
  rec = ParPredicate(in, rname);
  
  predicate = NULL;
  dbgmsg("*");
  if (ComToken == T_WHERE) {
    predicate = ParPredicate(in, rname);
  }
  if (predicate != NULL) {
    HakamaMergeRecord(rec, predicate, 0);
    HakamaDestroyRecord(predicate);
    predicate = NULL;
  }
  
  dbgmsg("*");
  
  CHECK_TERMINATOR;
  if (!in->fError) {
    ret = NewSqlInfo(SQL_TYPE_UPDATE);
    ret->tname = tname;
    ret->rname = rname;
    ret->pname = pname;
    ret->rec = rec;
    ret->fields = NULL;
    ret->limit = 1;
    ret->offset = 0;
  } else {
    ret = NULL;
  }
LEAVE_FUNC;
  return (ret);
}

static	SqlInfo	*
ParInsert(
	CURFILE			*in)
{
  SqlInfo	*ret;
  char		*tablename = NULL;
  Fields	*fields = NULL;
  Fields	*values = NULL;  
  HakamaRecord	*rec = NULL;
  char		*p
    ,		*tname
    ,		*rname
    ,		*pname
    ,		*name;
  int i;
  char field_name[SIZE_LONGNAME + 1];
  
ENTER_FUNC;
  GetSymbol;
  if (ComToken != T_INTO) {
    Error("INTO missing");
  }
  GetName;
  if ((ComToken !=  T_SYMBOL)
      && (ComToken !=  T_STRING)) {
    Error("table name invalid");
  } else {
    name = ParName(in);
    if ((p = strchr(name,'.'))  !=  NULL  ) {
      tablename = StrDup(p + 1);
    } else {
      tablename = StrDup(name);
    }
    xfree(name);
  }
  dbgprintf("tablename = [%s]\n", tablename);
  dbgmsg("*");
  
  tname = StrDup(tablename);
  if ((p = strchr(tablename,'$')) !=  NULL) {
    *p = 0;
    rname = StrDup(tablename);
    pname = StrDup(p + 1);
  } else {
    rname = StrDup(tablename);
    pname = StrDup("");
  }
  xfree(tablename);

  if (ComToken != '(') {
    Error("name fields invalid");
  } else {
    GetName;
  }
  fields = ParItemsName(in, T_VALUES);
  if (ComToken != ')') {
    Error("name fields invalid");
  }
  
  GetSymbol;
  if (ComToken == T_VALUES) {
    GetName;
    if (ComToken != '(') {
      Error("value fields invalid");
    }
    GetSymbol;
    values = ParItemsValue(in);
    if (ComToken != ')') {
      Error("value fields invalid");
    }
  }
  if (values && fields && fields->n == values->n) {
    dbgprintf("fields => %d, values => %d", fields->n, values->n);
    rec = HakamaNewRecord();
    for (i = 0; i < fields->n; ++i) {
      sprintf(field_name, "%s.%s", tname, fields->name[i]);
      HakamaEnterValue(rec, field_name, values->name[i]);
    }
  } else {
    Error("syntax error");
  }
  GetName;
  
  dbgmsg("*");
  CHECK_TERMINATOR;
  if (!in->fError) {
    ret = NewSqlInfo(SQL_TYPE_INSERT);
    ret->tname = tname;
    ret->rname = rname;
    ret->pname = pname;
    ret->rec = rec;
    ret->fields = NULL;
    ret->limit = 0;
    ret->offset = 0;
  } else {
    ret = NULL;
  }
  
  if (fields) {
    DestroyFields(fields);
  }
  if (values) {
    DestroyFields(values);
  }
  
LEAVE_FUNC;
  return (ret);
}

static	SqlInfo	*
ParDelete(
	CURFILE			*in)
{
  SqlInfo	*ret;
  char		*tablename;
  HakamaRecord	*predicate;
  char		*p
    ,		*tname
    ,		*rname
    ,		*pname
    ,		*name;
  
ENTER_FUNC;
  GetSymbol;
  if (ComToken != T_FROM) {
    Error("FROM missing");
  }
  GetName;
  if ((ComToken !=  T_SYMBOL)
      && (ComToken !=  T_STRING)) {
    Error("table name invalid");
  } else {
    name = ParName(in);
    if (tablename  ==  NULL) {
      if ((p = strchr(name,'.'))  !=  NULL  ) {
	tablename = StrDup(p + 1);
      } else {
	tablename = StrDup(name);
      }
    }
    xfree(name);
  }
  dbgprintf("tablename = [%s]\n", tablename);
  dbgmsg("*");
  tname = StrDup(tablename);
  if ((p = strchr(tablename,'$')) !=  NULL) {
    *p = 0;
    rname = StrDup(tablename);
    pname = StrDup(p + 1);
  } else {
    rname = StrDup(tablename);
    pname = StrDup("");
  }
  xfree(tablename);
  
  predicate = NULL;
  if (ComToken == T_WHERE) {
    predicate = ParPredicate(in, rname);
  }
  dbgmsg("*");
  CHECK_TERMINATOR;
  if (!in->fError) {
    ret = NewSqlInfo(SQL_TYPE_DELETE);
    ret->tname = tname;
    ret->rname = rname;
    ret->pname = pname;
    ret->rec = predicate;
    ret->fields = NULL;
    ret->limit = 1;
    ret->offset = 0;
  } else {
    ret = NULL;
  }
LEAVE_FUNC;
  return (ret);
}



typedef	struct _ArrayDimension	{
	int		count;
	struct	_ArrayDimension	*next;
}	ArrayDimension;

static	ItemStruct	*
NewItem(
	int		type)
{
	ItemStruct	*ret;

ENTER_FUNC;
	ret = New(ItemStruct);
	ret->type = type;
	switch	(type) {
	  case	ITEM_TYPE_BYTE:
	  case	ITEM_TYPE_BINARY:
	  case	ITEM_TYPE_CHAR:
	  case	ITEM_TYPE_VARCHAR:
	  case	ITEM_TYPE_DBCODE:
	  case	ITEM_TYPE_TEXT:
		ret->body.CharData.length = 0;
		break;
	  case	ITEM_TYPE_NUMBER:
		ret->body.NumberData.length = 0;
		ret->body.NumberData.slength = 0;
		break;
	  case	ITEM_TYPE_RECORD:
		ret->body.RecordData.count = 0;
		ret->body.RecordData.members = NewNameHash();
		ret->body.RecordData.item = NULL;
		ret->body.RecordData.names = NULL;
		break;
	  case	ITEM_TYPE_ARRAY:
		ret->body.ArrayData.count = 0;
		ret->body.ArrayData.item = NULL;
		break;
	  case	ITEM_TYPE_INT:
	  case	ITEM_TYPE_FLOAT:
	  case	ITEM_TYPE_BOOL:
	  case	ITEM_TYPE_OBJECT:
	  case	ITEM_TYPE_TIMESTAMP:
	  case	ITEM_TYPE_DATE:
	  case	ITEM_TYPE_TIME:
		break;
	  default:
		xfree(ret);
		ret = NULL;
		break;
	}
LEAVE_FUNC;
	return	(ret);
}


static	ItemStruct	*
ParItemDefine(
	CURFILE	*in)
{
	size_t		size
	,			ssize;
	int			token;
	ItemStruct	*item
	,			*array;
	ArrayDimension	*next
	,				*curr;

ENTER_FUNC;
	item = NULL;
	switch	(GetSymbol) {
	  case	T_BYTE:
	  case	T_CHAR:
	  case	T_NUMBER:
	  case	T_DBCODE:
	  case	T_VARCHAR:
	  case	T_TEXT:
		token = ComToken;
		size = 0;
		ssize = 0;
		if		(  GetSymbol  ==  '('  ) {
			if		(  GetSymbol  ==  T_ICONST  ) {
				size = ComInt;
				if		(  token  ==  T_NUMBER  ) {
					switch	(GetSymbol) {
					  case	')':
						ssize = 0;
						break;
					  case	',':
						if		(  GetSymbol  ==  T_ICONST  ) {
							ssize = ComInt;
						} else {
							Error("field size invalud");
						}
						GetSymbol;
						break;
					}
				} else {
					GetSymbol;
				}
				if		(  ComToken  !=  ')'  ) {
					Error("not closed left paren");
				}
			} else {
				Error("invalid size definition");
			}
			GetSymbol;
		} else {
			size = 1;
		}
		if		(  !in->fError  ) {
			switch	(token) {
			  case	T_BYTE:
				item = NewItem(ITEM_TYPE_BYTE);
				break;
			  case	T_CHAR:
				item = NewItem(ITEM_TYPE_CHAR);
				break;
			  case	T_VARCHAR:
				item = NewItem(ITEM_TYPE_VARCHAR);
				break;
			  case	T_DBCODE:
				item = NewItem(ITEM_TYPE_DBCODE);
				break;
			  case	T_NUMBER:
				item = NewItem(ITEM_TYPE_NUMBER);
				break;
			  case	T_TEXT:
				item = NewItem(ITEM_TYPE_TEXT);
			  default:
				break;
			}
			switch(item->type) {
			  case	ITEM_TYPE_NUMBER:
				item->body.NumberData.length = size;
				item->body.NumberData.slength = ssize;
				break;
			  default:
				item->body.CharData.length = size;
				break;
			}
		} else {
			item = NULL;
		}
		break;
	  case	T_BINARY:
		item = NewItem(ITEM_TYPE_BINARY);
		GetSymbol;
		break;
	  case	T_INT:
		item = NewItem(ITEM_TYPE_INT);
		GetSymbol;
		break;
	  case	T_FLOAT:
		item = NewItem(ITEM_TYPE_FLOAT);
		GetSymbol;
		break;
	  case	T_TIMESTAMP:
		item = NewItem(ITEM_TYPE_TIMESTAMP);
		GetSymbol;
		break;
	  case	T_DATE:
		item = NewItem(ITEM_TYPE_DATE);
		GetSymbol;
		break;
	  case	T_TIME:
		item = NewItem(ITEM_TYPE_TIME);
		GetSymbol;
		break;
	  case	T_BOOL:
		item = NewItem(ITEM_TYPE_BOOL);
		GetSymbol;
		break;
	  case	T_OBJECT:
		item = NewItem(ITEM_TYPE_OBJECT);
		GetSymbol;
		break;
	  case	'{':
	  case	'(':
		item = NewItem(ITEM_TYPE_RECORD);
		GetName;
		ParItemDefines(in,item);
		break;
	  default:
		item = NULL;
		Error("not supported");
		break;
	}
	next = NULL;
	while	(  ComToken  ==  '['  ) {	
		curr = New(ArrayDimension);
		if		(  GetSymbol  == T_ICONST  ) {
			curr->count = ComInt;
			GetSymbol;
		} else {
			curr->count = 0;
		}
		curr->next = next;
		next = curr;
		if		(  ComToken  !=  ']'  ) {
			Error("unmatch lbracket");
			break;
		}
		GetSymbol;
	}
	for	( curr = next ; curr != NULL ; ) {
		array = NewItem(ITEM_TYPE_ARRAY);
		array->body.ArrayData.count = curr->count;
		array->body.ArrayData.item = item;
		next = curr->next;
		xfree(curr);
		curr = next;
		item = array;
	}
LEAVE_FUNC;
	return	(item);
}

static	void
ParItemDefines(
	CURFILE		*in,
	ItemStruct	*upper)
{
	ItemStruct	*item;
	char		name[SIZE_LONGNAME+1];
	char		*dname;

ENTER_FUNC;
	while	(  ComToken  ==  T_SYMBOL  ) {
		strcpy(name,ComSymbol);
		item = ParItemDefine(in);
		if		(  ComToken  ==  ';'  ) {
			GetName;
		} else {
			Error("; missing");
		}
		if		(  !in->fError  ) {
			upper->body.RecordData.item = (ItemStruct **)
				xrealloc(upper->body.RecordData.item,
						 sizeof(ItemStruct *) * (upper->body.RecordData.count + 1));
			upper->body.RecordData.names = (char **)
				xrealloc(upper->body.RecordData.names,
						 sizeof(char *) * (upper->body.RecordData.count + 1));
			upper->body.RecordData.item[upper->body.RecordData.count] = item;
			dname = StrDup(name);
			upper->body.RecordData.names[upper->body.RecordData.count] = dname;
			g_hash_table_insert(upper->body.RecordData.members,
								(gpointer)dname,
								(gpointer)(upper->body.RecordData.count + 1));
			upper->body.RecordData.count ++;
		}
	}
	if		(	(  !in->fError      )
			&&	(	(  ComToken  ==  '}'  )
				||	(  ComToken  ==  ')'  ) ) ) {
		GetSymbol;
		/*	OK	*/
	} else {
		dbgprintf("token = %d [%c]\n",ComToken,ComToken);
		Error("syntax error(invalid structure define(s))");
	}
LEAVE_FUNC;
}

static	ItemStruct	*
ParseRecord(
	CURFILE	*in)
{
	ItemStruct	*ret;

ENTER_FUNC;
	ret = NULL;
	if		(  ComToken  ==  T_SYMBOL  ) {
		GetSymbol;
	}
	if		(  ComToken  ==  '{'  ) {
		ret = NewItem(ITEM_TYPE_RECORD);
		GetName;
		ParItemDefines(in,ret);
		if		(  in->fError  ) {
			ret = NULL;
		}
	} else {
		ret = NULL;
		Error("syntax error");
	}
LEAVE_FUNC;
	return	(ret);
}

extern	void
DestroyItemStruct(
	ItemStruct	*item)
{
	int		i;

ENTER_FUNC;
	if		(  item  !=  NULL  ) {
		dbgprintf("type = %02X\n",item->type);
		switch	(item->type) {
		  case	ITEM_TYPE_ARRAY:
			DestroyItemStruct(item->body.ArrayData.item);
			break;
		  case	ITEM_TYPE_RECORD:
			for	( i = 0 ; i < item->body.RecordData.count ; i ++ ) {
				DestroyItemStruct(item->body.RecordData.item[i]);
				xfree(item->body.RecordData.names[i]);
			}
			g_hash_table_destroy(item->body.RecordData.members);
			xfree(item->body.RecordData.names);
			xfree(item->body.RecordData.item);
			break;
		  case	ITEM_TYPE_BYTE:
		  case	ITEM_TYPE_BINARY:
		  case	ITEM_TYPE_CHAR:
		  case	ITEM_TYPE_VARCHAR:
		  case	ITEM_TYPE_DBCODE:
		  case	ITEM_TYPE_TEXT:
		  case	ITEM_TYPE_NUMBER:
		  case	ITEM_TYPE_OBJECT:
		  default:
			break;
		}
		xfree(item);
	}
LEAVE_FUNC;
}

extern	void
DestroySqlInfo(
	SqlInfo	*info)
{
ENTER_FUNC;
 if (info == NULL) {
   return;
 }
	if		(  info->tname  !=  NULL  ) {
		xfree(info->tname);
	}
	if		(  info->rname  !=  NULL  ) {
		xfree(info->rname);
	}
	if		(  info->pname  !=  NULL  ) {
		xfree(info->pname);
	}
	if		(  info->rec  !=  NULL  ) {
		HakamaDestroyRecord(info->rec);
	}
	if		(  info->record  !=  NULL  ) {
		DestroyItemStruct(info->record);
	}
	xfree(info);
LEAVE_FUNC;	
}

static	SqlInfo	*
Parser(
	CURFILE	*in)
{
	SqlInfo	*sql;

ENTER_FUNC;
	sql = NULL;
	while	(  GetSymbol  !=  T_EOF  ) {
		switch	(ComToken) {
		  case	T_SELECT:
			sql = ParSelect(in);
			break;
		  case	T_UPDATE:
			sql = ParUpdate(in);
			break;
		  case	T_INSERT:
			sql = ParInsert(in);
			break;
		  case	T_DELETE:
			sql = ParDelete(in);
			break;
		  default:
			sql = NewSqlInfo(SQL_TYPE_RECORD);
			sql->record = ParseRecord(in);
			break;
		}
		if		(  ComToken  ==  T_EOF  )	break;
	}
	if		(  sql  !=  NULL  ) {
		dbgprintf("type  = %d",sql->type);
		dbgprintf("tname = [%s]",sql->tname);
		dbgprintf("rname = [%s]",sql->rname);
		dbgprintf("pname = [%s]",sql->pname);
		dbgprintf("limit = [%d]",sql->limit);
#ifdef DEBUG		
		HakamaDumpRecord(sql->rec);
		DumpItemStruct(sql->record);
#endif
	}
LEAVE_FUNC;
	return	(sql);
}

static	GHashTable	*Reserved = NULL;

extern	SqlInfo	*
CompileSQL(
	char	*sql)
{
  CURFILE	*in;
  SqlInfo	*ret;

ENTER_FUNC;
 
 dbgmsg(sql);

 if (setjmp(compiler_env)) {
   dbgmsg("compile error");
LEAVE_FUNC;   
   return NULL;
 }
 in = New(CURFILE);
 in->body = sql;
 in->back = -1;
 in->size = -1;
 in->pos = 0;
 in->cLine = 1;
 in->fError = FALSE;
 in->Symbol = NULL;
 in->reserved = Reserved;
 ret = Parser(in);
 
LEAVE_FUNC;
 return	(ret);
}

extern	ItemStruct	*
CompileRecord(
	char	*code)
{
  CURFILE	*in;
  ItemStruct	*ret;

ENTER_FUNC;
 if (setjmp(compiler_env)) {
   dbgmsg("compile error");   
LEAVE_FUNC;   
   return NULL;
 }

 in = New(CURFILE);
 in->body = code;
 in->back = -1;
 in->size = -1;
 in->pos = 0;
 in->cLine = 1;
 in->fError = FALSE;
 in->Symbol = NULL;
 in->reserved = Reserved;
 
 GetSymbol;
 ret = ParseRecord(in);
LEAVE_FUNC;
 return	(ret);
}

static	TokenTable	tokentable[] = {
	{	"CREATE",		T_CREATE	},
	{	"DROP",			T_DROP	 	},

	{	"INSERT",		T_INSERT	},
	{	"DELETE",		T_DELETE 	},
	{	"UPDATE",		T_UPDATE 	},
	{	"SELECT",		T_SELECT 	},
	{	"SET",			T_SET	 	},
	{	"DECLARE",		T_DECLARE 	},
	{	"CLOSE",		T_CLOSE 	},
	{	"FETCH",		T_FETCH 	},

	{	"START",		T_START 	},
	{	"COMMIT",		T_COMMIT 	},
	{	"ABORT",		T_ABORT 	},

	{	"TABLE",		T_TABLE 	},
	{	"INDEX",		T_INDEX		},
	{	"DICTIONARY",	T_DICTIONARY},
	{	"LOCALE",		T_LOCALE	},
	{	"TRIGGER",		T_TRIGGER	},
	{	"CURSOR",		T_CURSOR	},
	{	"LIMIT",		T_LIMIT		},
	{	"OFFSET",		T_OFFSET	},

	{	"ON",			T_ON 		},
	{	"USING",		T_USING		},
	{	"OPTION",		T_OPTION	},
	{	"INTO",			T_INTO		},
	{	"VALUES",		T_VALUES	},
	{	"FROM",			T_FROM		},
	{	"KEY",			T_KEY		},
	{	"WHERE",		T_WHERE		},
	{	"FOR",			T_FOR		},
	{	"AS",			T_AS		},
	{	"ORDER",		T_ORDER		},
	{	"BY",			T_BY		},
	{	"FORWARD",		T_FORWARD	},
	{	"BACKWARD",		T_BACKWARD	},
	{	"RELATIVE",		T_RELATIVE	},
	{	"IN",			T_IN		},
	{	"DESC",			T_DESC		},
	{	"ASC",			T_ASC		},

	{	"ALL",			T_ALL		},
	{	"DISTINCT",		T_DISTINCT	},

	{	"CROSS",		T_CROSS		},
	{	"INNER",		T_INNER		},
	{	"OUTER",		T_OUTER		},
	{	"JOIN",			T_JOIN		},
	{	"LEFT",			T_LEFT		},
	{	"RIGHT",		T_RIGHT		},
	{	"FULL",			T_FULL		},

	{	"NULL",			T_NULL		},

	{	"AND",			T_AND		},
	{	"OR",			T_OR		},
	{	"NOT",			T_NOT		},

	{	"LIKE",			T_LIKE		},
	{	"BETWEEN",		T_BETWEEN	},

	{	"bool",			T_BOOL		},
	{	"byte",			T_BYTE		},
	{	"binary",		T_BINARY	},
	{	"char",			T_CHAR		},
	{	"varchar",		T_VARCHAR	},
	{	"float",		T_FLOAT	},
	{	"timestamp",	T_TIMESTAMP},
	{	"date",			T_DATE		},
	{	"time",			T_TIME		},
	{	"enum",			T_ENUM		},
	{	"int",			T_INT		},
	{	"integer",		T_INT		},
	{	"number",		T_NUMBER	},
	{	"text",			T_TEXT		},
	{	"object",		T_OBJECT	},
	{	"dbcode",		T_DBCODE	},
	{	"primary",		T_PRIMARY	},
	{	"key",			T_KEY		},

	{	""				,0			}
};

extern	void
CompileInit(void)
{
ENTER_FUNC;
	if		(  Reserved  ==  NULL  ) {
		Reserved = LexInit(tokentable);
	}
LEAVE_FUNC;
}

extern	void
DumpItemStruct(
	ItemStruct	*item)
{
	int		i;

	if		(  item  ==  NULL  )	{
		printf("null value\n");
	} else {
		switch	(item->type) {
		  case	ITEM_TYPE_INT:
			printf("integer\n");
			fflush(stdout);
			break;
		  case	ITEM_TYPE_FLOAT:
			printf("float\n");
			fflush(stdout);
			break;
		  case	ITEM_TYPE_BOOL:
			printf("Bool\n");
			fflush(stdout);
			break;
		  case	ITEM_TYPE_CHAR:
			printf("char(%d)\n", item->body.CharData.length);
			fflush(stdout);
			break;
		  case	ITEM_TYPE_VARCHAR:
			printf("varchar(%d)\n", item->body.CharData.length);
			fflush(stdout);
			break;
		  case	ITEM_TYPE_NUMBER:
			printf("number(%d,%d)\n",
				   (int)item->body.NumberData.length,
				   (int)item->body.NumberData.slength);
			fflush(stdout);
			break;
		  case	ITEM_TYPE_TEXT:
			printf("text(%d)\n", item->body.CharData.length);
			fflush(stdout);
			break;
		  case	ITEM_TYPE_BINARY:
			printf("binary(%d)\n", item->body.CharData.length);
			fflush(stdout);
			break;
		  case	ITEM_TYPE_TIMESTAMP:
			printf("datetime\n");
			break;
		  case	ITEM_TYPE_DATE:
			printf("date\n");
			break;
		  case	ITEM_TYPE_TIME:
			printf("time\n");
			break;
		  case	ITEM_TYPE_ARRAY:
			printf("array size = %d\n",
				   (int)item->body.ArrayData.count);
			fflush(stdout);
			DumpItemStruct(item->body.ArrayData.item);
			break;
		  case	ITEM_TYPE_RECORD:
			printf("record members = %d\n",(int)item->body.RecordData.count);
			fflush(stdout);
			for	( i = 0 ; i < (int)item->body.RecordData.count ; i ++ ) {
				printf("%s:",item->body.RecordData.names[i]);
				DumpItemStruct(item->body.RecordData.item[i]);
			}
			printf("--\n");
			break;
		  default:
			break;
		}
	}
}

