/*
#define	CONSOLE_DEBUG
#define	VERSION		"1.4.3"
#define	DEBUG
#define	TRACE
*/

#define	VERSION		"1.5.1"

#include	"socket.h"
#include	"hakama.h"
#include	"libs.h"
#include	"compile.h"
#include	"iconv.h"
#include	"debug.h"
#include	<stdio.h>
#include	<ctype.h>
#include	<tchar.h>
#include        <limits.h>

#define	MSG_EXEC		"Exec: "
#define	MSG_SCHEMA		"Schema: "
#define	MSG_PATHTABLES	"PathTables: "

#if	0
static	size_t
Read(
	SOCKET	sd,
	char	*buff,
	size_t	size)
{
	char	*p = buff;
	ssize_t	count;
	int		ret;

	ret = size;
	while	(  size  >  0  ) {
		if		(  ( count = recv(sd,p,size,0) )  >  0  ) {
			size -= count;
			p += count;
		} else {
			ret = -1;
			break;
		}
	}
	return	(ret);
}
#endif

static	int
ParseVersion(
	char	*str)
{
	int		major;
	int		minor;
	int		micro;

	major = 0;
	minor = 0;
	micro = 0;

	while	(	(  *str  !=  0    )
			&&	(  isdigit(*str)  ) )	{
		major = major * 10 + ( *str - '0' );
		str ++;
	}
	if		(  *str  !=  0  )	str ++;
	while	(	(  *str  !=  0    )
			&&	(  isdigit(*str)  ) )	{
		minor = minor * 10 + ( *str - '0' );
		str ++;
	}
	if		(  *str  !=  0  )	str ++;
	while	(	(  *str  !=  0    )
			&&	(  isdigit(*str)  ) )	{
		micro = micro * 10 + ( *str - '0' );
		str ++;
	}
	dbgprintf("%d.%d.%d",major,minor,micro);
	return	(major * 10000 + minor * 100 + micro);
}

static	int
RecvChar(
	SOCKET	sd)
{
	char	data;
	int		ret;

	if		(  recv(sd,&data,sizeof(data),0)  !=  SOCKET_ERROR  ) {
		ret = data;
	} else {
		ret = -1;
	}
	return	(ret);
}

static	Bool
RecvStringDelim(
	SOCKET	sd,
	size_t	size,
	char	*str)
{
	Bool	rc;
	int		c;
	char	*p;

	p = str;
	while	(	(  ( c = RecvChar(sd) )  >=  0     )
			&&	(  c                     !=  '\n'  ) )	{
		*p ++ = c;
		size --;
		if		(  size  ==  0  )	break;
	}
	*p = 0;
	if		(  c  >=  0  ) {
		StringChop(str);
		rc = TRUE;
	} else {
		rc = FALSE;
	}
	return	(rc);
}

static	int
Send(
	SOCKET	sd,
	void	*buff,
	size_t	size)
{
	char	*p = buff;
	ssize_t	count;
	int		ret;

	ret = size;
	while	(  size  >  0  ) {
		if		(  ( count = send(sd,p,size,0) )  !=  SOCKET_ERROR  ) {
			size -= count;
			p += count;
		} else {
			ret = -1;
			break;
		}
	}
	return	(ret);
}

static	Bool
SendStringDelim(
	SOCKET	sd,
	char	*str)
{
	size_t	size;
	Bool	ret;

	if		(   str  !=  NULL  ) { 
		size = strlen(str);
	} else {
		size = 0;
	}
	if		(  size  >  0  ) {
		ret = Send(sd,str,size) >= 0 ? TRUE : FALSE;
	} else {
		ret = TRUE;
	}
	return	(ret);
}

/*
 *	main functions
 */


extern	Hakama	*
HakamaConnect(
	char	*host,
	char	*port,
	char	*user,
	char	*pass)
{
	Hakama	*ret;
	char	buf[SIZE_BUFF+1];
	SOCKET	sock;
	char	*p;

ENTER_FUNC;
	ret = NULL;
	if		(  ( sock = ConnectIP_Socket(port,SOCK_STREAM,host) )  !=  INVALID_SOCKET  ) {
		sprintf(buf,"%s %s %s stringe\n",VERSION,user,pass);
		dbgprintf("send = [%s]",buf);
		SendStringDelim(sock,buf);
		RecvStringDelim(sock,SIZE_BUFF,buf);
		dbgprintf("recv = [%s]",buf);
		if		(  strlcmp(buf,"ERROR")  ==  0  ) {
			close(sock);
		} else {
			ret = New(Hakama);
			ret->sd = sock;
			ret->count = 0;
			ret->lastrc = 0;
			ret->info = NULL;
			if		(  ( p = strchr(buf,';' ) )  !=  NULL  ) {
				ret->version = ParseVersion(p+1);
			} else {
				ret->version = 10403;
			}
		}
	}
LEAVE_FUNC;
	return	(ret);
}

static	void
_SendData(
	char			*name,
	HakamaKeyValue	*rec,
	Hakama			*hakama)
{
	char	*buf;

	SendStringDelim(hakama->sd,name);
	SendStringDelim(hakama->sd,": ");
	buf = (char *)xmalloc(strlen(rec->data) * 3);
	dbgprintf("[%s]", rec->data);
	EncodeStringURL(buf,rec->data);
	dbgprintf("[%s]", buf);
	SendStringDelim(hakama->sd,buf);
	SendStringDelim(hakama->sd,"\n");
	xfree(buf);
}

static	void
ExecData(
	Hakama			*hakama,
	HakamaRecord	*rec)
{
ENTER_FUNC;
	if		(  rec  !=  NULL  ) {
		g_hash_table_foreach(rec->item,(GHFunc)_SendData,hakama);
	}
	SendStringDelim(hakama->sd,"\n");
LEAVE_FUNC;
}

static	int
GetEvent(
	Hakama	*hakama)
{
	char	buf[SIZE_BUFF+1];
	int		rc = 0;
	char	*p
		,	*q;

ENTER_FUNC;
	if		(  RecvStringDelim(hakama->sd,SIZE_BUFF,buf)  ) {
		dbgprintf("[%s]",buf);
		if		(  strlcmp(buf,MSG_EXEC)  ==  0  ) {
			p = buf + strlen(MSG_EXEC);
			if		(  ( q = strchr(p,':') )  !=  NULL  ) {
				*q = 0;
				rc = atoi(p);
				hakama->count = atoi(q+1);
				dbgprintf("rc = %s,%d", p, rc);
			} else {
				rc = atoi(p);
				hakama->count = 1;
				dbgprintf("rc = %d", rc);
			}
		} else {
			rc = -1;
		}
	} else {
		rc = -1;
	}
	if		(  rc  ==  -1  ) {
		close(hakama->sd);
		hakama->count = 0;
	}
	hakama->lastrc = rc;
	
LEAVE_FUNC;
	return	(rc);
}

extern	int
HakamaExec(
	Hakama	*hakama,
	char	*func)
{
	char	buf[SIZE_BUFF+1];
	int		rc;

ENTER_FUNC;
	dbgprintf("%s",func);
	sprintf(buf,"%s%s\n\n",MSG_EXEC,func);
	if		(  SendStringDelim(hakama->sd,buf)  ) {
		rc = GetEvent(hakama);
		if		(  hakama->version  <  10500  ) {
			SendStringDelim(hakama->sd,"\n");
		}
	} else {
		rc = -1;
	}
LEAVE_FUNC;
	return	rc;
}

extern	void
HakamaDisconnect(
	Hakama	**hakamap)
{
	Hakama *hakama;
	if (!hakamap) {
	  return;
	}
	hakama = *hakamap;

	HakamaExec(hakama,"DBDISCONNECT");
	SendStringDelim(hakama->sd,"End\n");
	close(hakama->sd);

	if (hakama->info) {
	  DestroySqlInfo(hakama->info);
	  hakama->info = NULL;
	}
	xfree(hakama->input_enc.from);
	xfree(hakama->input_enc.to);
	xfree(hakama->output_enc.from);
	xfree(hakama->output_enc.to);
	
	xfree(hakama);
	*hakamap = NULL;
}

extern	HakamaRecord	*
HakamaNewRecord(void)
{
	HakamaRecord	*ret;

	ret = New(HakamaRecord);
	ret->count = 0;
	ret->item = NewNameHash();
	ret->vals = NULL;
	return	(ret);
}

extern	HakamaRecords	*
HakamaNewRecords(void)
{
	HakamaRecords	*ret;

	ret = New(HakamaRecords);
	ret->rec = NULL;
	ret->count = 0;
	return	(ret);
}

extern	void
HakamaEnterValue(
	HakamaRecord	*rec,
	char			*name,
	char			*value)
{
	HakamaKeyValue	*val;

	val = New(HakamaKeyValue);
	val->name = StrDup(name);
	val->data = StrDup(value);
	rec->vals = (HakamaKeyValue **)xrealloc(rec->vals,sizeof(HakamaKeyValue *) * ( rec->count + 1 ));
	rec->vals[rec->count] = val;
	rec->count ++;
	g_hash_table_insert(rec->item,val->name,val);
}

extern	char	*
HakamaGetValue(
	HakamaRecord	*rec,
	char			*name)
{
	HakamaKeyValue	*val;
	char	*ret;

	if		(  ( val = (HakamaKeyValue *)g_hash_table_lookup(rec->item,name) )
			   !=  NULL  ) {
		ret = val->data;
	} else {
		ret = NULL;
	}
	return	(ret);
}

extern	void
HakamaDumpRecord(
	HakamaRecord	*rec)
{
	int		i;
ENTER_FUNC;
	if		(  rec  !=  NULL  ) {
		for	( i = 0 ; i < rec->count ; i ++ ) {
			if		(  rec->vals[i]  !=  NULL  ) {
				printf("%s: %s\n",rec->vals[i]->name,rec->vals[i]->data);
			}
		}
	}
LEAVE_FUNC;
}

extern	void
HakamaDumpRecords(
	HakamaRecords	*recs)
{
	int		i;
ENTER_FUNC;
	if		(  recs  !=  NULL  ) {
		for	( i = 0 ; i < recs->count ; i ++ ) {
			if		(  recs->rec[i]  !=  NULL  ) {
				HakamaDumpRecord(recs->rec[i]);
			}
		}
	}
LEAVE_FUNC;
}

extern	void
HakamaAppendRecord(
	HakamaRecords	*recs,
	HakamaRecord	*rec)
{
	if		(  recs  !=  NULL  ) {
		recs->rec = (HakamaRecord **)xrealloc(recs->rec,
											  sizeof(HakamaRecord *) * ( recs->count + 1 ));
		recs->rec[recs->count] = rec;
		recs->count ++;
	}
}

#define	LastRecord(recs)		((recs)->rec[(recs)->count-1])

static	HakamaRecords	*
HakamaGetValues(
	Hakama	*hakama,
	char	*name,
	char	*rename,
	int		limit)
{
	char	buf[SIZE_BUFF+1]
	  ,	*buf1
	  ,	*buf2;
	  ;
	HakamaRecords	*ret;
	HakamaRecord	*item;
	char	*p
		,	*q
		,	*r
		,	*rname;
	int read_count = 0;

ENTER_FUNC;
	if		(  hakama->count  ==  0  ) {
		strcpy(buf,"\n");
	} else {
		if		(  limit  ==  1  ) {
			sprintf(buf,"%s\n",name);
		} else {
			sprintf(buf,"%s;%d\n",name,limit);
		}
	}
	if		(  !SendStringDelim(hakama->sd,buf)  )	return	NULL;
	if		(  hakama->count  ==  0  )	return	NULL;


	dbgprintf("count => [%d]", hakama->count);
	if (limit > 0 && limit < hakama->count) {
	  hakama->count = limit;
	}
	ret = HakamaNewRecords();
	item = NULL;
	buf1 = (char *)xmalloc(SIZE_LARGEBUFF+1);
	buf2 = (char *)xmalloc(SIZE_LARGEBUFF+1);
	while	(  RecvStringDelim(hakama->sd,SIZE_LARGEBUFF,buf1)  )	{
	  dbgprintf("recv [%s] count [%d] read_count [%d]",buf1,hakama->count, read_count);
		if		(  strlen(buf1)  ==  0  ) {
			if		(  item  ==  NULL  )		break;
			
			HakamaConvCharsetRecord(hakama->input_enc, item);
			HakamaAppendRecord(ret,item);
			++read_count;
			item = NULL;
			if (hakama->version < 10501) {
			  if		(  hakama->count  ==  1  )	break;
			}
		} else {
			if		(  item  ==  NULL  ) {
				item = HakamaNewRecord();
			}
			if		(  ( p = strchr(buf1,':') )  !=  NULL  ) {
				*p = 0;
				if		(  ( q = strchr(buf1,';') )  !=  NULL  ) {
					*q = 0;
					if		(  rename  ==  NULL  ) {
						rname = StrDup(buf1);
					} else {
						if		(  ( r = strchr(buf1,'.') )  !=  NULL  ) {
							*r = 0;
							if		(  strcmp(buf1,name)  ==  0  ) {
								strcpy(buf,rename);
								q = buf + strlen(rename);
								*q ++ = '.';
								strcpy(q,r+1);
								rname = StrDup(buf);
							} else {
								*r = '.';
								rname = StrDup(buf1);
							}
						} else {
							rname = StrDup(buf1);
						}
					}
					memset(buf2, 0, sizeof(buf2));
					DecodeStringURL((byte *)buf2,p+2);
					HakamaEnterValue(item,rname,buf2);
				}
			}
		}
	}
	xfree(buf1);
	xfree(buf2);
	dbgmsg("send terminater");
	SendStringDelim(hakama->sd,"\n");

	dbgprintf("read_count => [%d]", read_count);
	hakama->count = read_count;
LEAVE_FUNC;
	return	(ret);
}

extern	HakamaRecords	*
HakamaExecRecord(
	Hakama	*hakama,
	char	*func,
	char	*rname,
	char	*pname,
	char	*rename,
	HakamaRecord	*rec,
	int		limit
	)
{
	char	buf[SIZE_BUFF+1];
	int rc;
	HakamaRecords	*ret;
	
ENTER_FUNC;
	if		(  limit  ==  1  ) {
		sprintf(buf,"%s%s:%s:%s\n",MSG_EXEC,func,rname,pname);
	} else {
		sprintf(buf,"%s%s:%s:%s:%d\n",MSG_EXEC,func,rname,pname,limit);
	}
	
	dbgprintf("send [%s]",buf);
	if		(  !SendStringDelim(hakama->sd,buf)  )	return	NULL;
	ExecData(hakama,rec);
	rc = GetEvent(hakama);
	dbgprintf("count => [%d]", hakama->count);
	if (rc < 0) {
	  ret = NULL;
	} else {
	  if	(((  hakama->version  >=  10500  )
		  && (hakama->count    ==  0      ))
		 || !hakama->isselect)  {
	    ret = HakamaNewRecords();
	  } else {
	    ret = HakamaGetValues(hakama,rname,rename,limit);
	  }
	}

LEAVE_FUNC;
	return	(ret);
}

extern	HakamaRecords	*
HakamaExecSql(
	Hakama	*hakama,
	char	*sql,
	int     *compile_error)
{
	SqlInfo	*info = NULL;
	char	*rename;
	HakamaRecords *ret;
	
ENTER_FUNC;
        *compile_error = 0;
	CompileInit();
	ret = NULL;
	if (hakama->info) {
	  DestroySqlInfo(info);
	  hakama->info = NULL;
	}
	if		(  ( info = CompileSQL(sql) )  !=  NULL  ) {
		if		(	(  info->pname   !=  NULL  )
				&&	(  *info->pname  !=  0     ) ) {
			rename = info->tname;
		} else {
			rename = NULL;
		}
		hakama->isselect = 0;

		if (info->rec) {
		  HakamaConvCharsetRecord(hakama->output_enc, info->rec);
		}
		
		switch	(info->type) {
		  case	SQL_TYPE_SELECT:
	    		hakama->isselect = 1;
			ret = HakamaExecRecord(hakama,"DBSELECT",
					      info->rname,info->pname,rename,info->rec,info->limit);
			break;
		  case	SQL_TYPE_DELETE:
			ret = HakamaExecRecord(hakama,"DBDELETE",
					      info->rname,info->pname,rename,info->rec,info->limit);
			break;
		  case	SQL_TYPE_UPDATE:
			ret = HakamaExecRecord(hakama,"DBUPDATE",
					      info->rname,info->pname,rename,info->rec,info->limit);
			break;
		  case	SQL_TYPE_INSERT:
			ret = HakamaExecRecord(hakama,"DBINSERT",
					      info->rname,info->pname,rename,info->rec,info->limit);
			break;
		  default:
			break;
		}
	} else {
	  dbgmsg("compile error");
	  *compile_error = 1;
	}
	hakama->info = info;
LEAVE_FUNC;
	return	(ret);
}

static	HakamaList	*
HakamaNewList(void)
{
	HakamaList	*ret;

	ret = New(HakamaList);
	ret->count = 0;
	ret->item = NULL;
	return	(ret);

}

extern	void
HakamaDestroyList(
	HakamaList	*list)
{
	int		i;

	if		(  list  !=  NULL  ) {
		for	( i = 0 ; i < list->count ; i ++ ) {
			if		(  list->item[i]  !=  NULL  ) {
				xfree(list->item[i]);
			}
		}
		xfree(list);
	}
}

static	HakamaList	*
HakamaGetList(
	Hakama	*hakama)
{
	HakamaList	*ret;
	char	buf[SIZE_BUFF+1];
ENTER_FUNC;
	ret = HakamaNewList();
	while	(  RecvStringDelim(hakama->sd,SIZE_BUFF,buf)  )	{
		dbgprintf("recv [%s]",buf);
		if		(  strlen(buf)  ==  0  )	break;
		ret->item = (char **)xrealloc(ret->item,
										   sizeof(char *) * ( ret->count + 1));
		ret->item[ret->count] = StrDup(buf);
		ret->count ++;
	}
	if		(  ret->count  ==  0  ) {
		HakamaDestroyList(ret);
		ret = NULL;
	}
LEAVE_FUNC;
	return	(ret);
}
extern	HakamaList	*
HakamaGetPathTables(
	Hakama	*hakama,
	char	*table)
{
	char	buf[SIZE_BUFF+1];
	HakamaList	*ret;

ENTER_FUNC;
	if		(	(  table   ==  NULL  )
			||	(  *table  ==  0     ) ) {
		sprintf(buf,"%s\n",MSG_PATHTABLES);
	} else {
		sprintf(buf,"%s%s\n",MSG_PATHTABLES,table);
	}
	dbgprintf("send [%s]",buf);
	if		(  !SendStringDelim(hakama->sd,buf)  )	return	NULL;
	ret = HakamaGetList(hakama);
LEAVE_FUNC;
	return	(ret);
}

static	void
PutDim(
	HakamaRecord	*rec,
	int		*dim,
	int		alevel)
{
	char	*p;
	char	buff[SIZE_LONGNAME+1];
	int		i;

	p = buff;
	*p = 0;
	for	( i = alevel - 1 ; i >= 0 ; i -- ) {
		if		(  dim[i]  ==  0  ) {
			p += sprintf(p,"[]");
		} else {
			p += sprintf(p,"[%d]",dim[i]);
		}
	}
	if		(  *buff  !=  0  ) {
		HakamaEnterValue(rec,"array",buff);
	}
}

static	void
PutItemName(
	HakamaRecord	*rec,
	char	**rname,
	int		level)
{
	char	*p;
	char	buff[SIZE_LONGNAME+1];
	int		i;

	p = buff;
	if		(  level  >  1  ) {
		for	( i = 0 ; i < level - 1 ; i ++ ) {
			p += sprintf(p,"%s_",rname[i]);
		}
	}
	sprintf(p,"%s",rname[level-1]);
	HakamaEnterValue(rec,"name",buff);
}

static	void
ConvertRecordBody(
	HakamaRecords	*recs,
	ItemStruct		*item,
	int				level,
	int				alevel,
	char			**rname,
	int				*dim)
{
	char	buff[SIZE_LONGNAME+1];
	int		i;

ENTER_FUNC;
	if		(  item  !=  NULL  )	{
		switch	(item->type) {
		  case	ITEM_TYPE_INT:
			dbgmsg("ITEM_TYPE_INT");
			HakamaAppendRecord(recs,HakamaNewRecord());
			PutItemName(LastRecord(recs),rname,level);
			PutDim(LastRecord(recs),dim,alevel);
			HakamaEnterValue(LastRecord(recs),"type","integer");
			break;
		  case	ITEM_TYPE_FLOAT:
			dbgmsg("ITEM_TYPE_FLOAT");
			HakamaAppendRecord(recs,HakamaNewRecord());
			PutItemName(LastRecord(recs),rname,level);
			PutDim(LastRecord(recs),dim,alevel);
			HakamaEnterValue(LastRecord(recs),"type","float");
			break;
		  case	ITEM_TYPE_BOOL:
			dbgmsg("ITEM_TYPE_BOOL");
			HakamaAppendRecord(recs,HakamaNewRecord());
			PutItemName(LastRecord(recs),rname,level);
			PutDim(LastRecord(recs),dim,alevel);
			HakamaEnterValue(LastRecord(recs),"type","char(1)");
			break;
		  case	ITEM_TYPE_CHAR:
			dbgmsg("ITEM_TYPE_CHAR");
			HakamaAppendRecord(recs,HakamaNewRecord());
			PutItemName(LastRecord(recs),rname,level);
			PutDim(LastRecord(recs),dim,alevel);
			sprintf(buff,"char(%d)", item->body.CharData.length);
			HakamaEnterValue(LastRecord(recs),"type",buff);
			break;
		  case	ITEM_TYPE_VARCHAR:
			dbgmsg("ITEM_TYPE_VARCHAR");
			HakamaAppendRecord(recs,HakamaNewRecord());
			PutItemName(LastRecord(recs),rname,level);
			PutDim(LastRecord(recs),dim,alevel);
			sprintf(buff,"varchar(%d)", item->body.CharData.length);
			HakamaEnterValue(LastRecord(recs),"type",buff);
			break;
		  case	ITEM_TYPE_NUMBER:
			dbgmsg("ITEM_TYPE_NUMBER");
			HakamaAppendRecord(recs,HakamaNewRecord());
			PutItemName(LastRecord(recs),rname,level);
			PutDim(LastRecord(recs),dim,alevel);
			sprintf(buff,"number(%d,%d)\n",
					(int)item->body.NumberData.length,
					(int)item->body.NumberData.slength);
			HakamaEnterValue(LastRecord(recs),"type",buff);
			break;
		  case	ITEM_TYPE_TEXT:
			dbgmsg("ITEM_TYPE_TEXT");
			HakamaAppendRecord(recs,HakamaNewRecord());
			PutItemName(LastRecord(recs),rname,level);
			PutDim(LastRecord(recs),dim,alevel);
			HakamaEnterValue(LastRecord(recs),"type","text");
			break;
		  case	ITEM_TYPE_BINARY:
			dbgmsg("ITEM_TYPE_BINARY");
			HakamaAppendRecord(recs,HakamaNewRecord());
			PutItemName(LastRecord(recs),rname,level);
			PutDim(LastRecord(recs),dim,alevel);
			HakamaEnterValue(LastRecord(recs),"type","bytea");
			break;
		  case	ITEM_TYPE_TIMESTAMP:
			dbgmsg("ITEM_TYPE_TIMESTAMP");
			HakamaAppendRecord(recs,HakamaNewRecord());
			PutItemName(LastRecord(recs),rname,level);
			PutDim(LastRecord(recs),dim,alevel);
			HakamaEnterValue(LastRecord(recs),"type","timestamp");
			break;
		  case	ITEM_TYPE_DATE:
			dbgmsg("ITEM_TYPE_DATE");
			HakamaAppendRecord(recs,HakamaNewRecord());
			PutItemName(LastRecord(recs),rname,level);
			PutDim(LastRecord(recs),dim,alevel);
			HakamaEnterValue(LastRecord(recs),"type","date");
			break;
		  case	ITEM_TYPE_TIME:
			dbgmsg("ITEM_TYPE_TIME");
			HakamaAppendRecord(recs,HakamaNewRecord());
			PutItemName(LastRecord(recs),rname,level);
			PutDim(LastRecord(recs),dim,alevel);
			HakamaEnterValue(LastRecord(recs),"type","time");
			break;
		  case	ITEM_TYPE_OBJECT:
			dbgmsg("ITEM_TYPE_OBJECT");
			HakamaAppendRecord(recs,HakamaNewRecord());
			PutItemName(LastRecord(recs),rname,level);
			PutDim(LastRecord(recs),dim,alevel);
			HakamaEnterValue(LastRecord(recs),"type","oid");
			break;
		  case	ITEM_TYPE_ARRAY:
			dbgmsg("ITEM_TYPE_ARRAY");
			dim[alevel] = item->body.ArrayData.count;
			ConvertRecordBody(recs,item->body.ArrayData.item,level,(alevel+1),rname,dim);
			break;
		  case	ITEM_TYPE_RECORD:
			dbgmsg("ITEM_TYPE_RECORD");
			for	( i = 0 ; i < (int)item->body.RecordData.count ; i ++ ) {
				rname[level] =  item->body.RecordData.names[i];
				ConvertRecordBody(recs,item->body.RecordData.item[i],
								 (level+1),alevel,rname,dim);
			}
			break;
		  default:
			break;
		}
	}
LEAVE_FUNC;
}

extern	void
ConvertRecord(
	HakamaRecords	*recs,
	ItemStruct		*item)
{
	char	*rname[20];
	int		dim[20];

ENTER_FUNC;
	ConvertRecordBody(recs,item,0,0,rname,dim);
LEAVE_FUNC;
}

extern	HakamaRecords	*
HakamaGetPathSchema(
	Hakama	*hakama,
	char	*name)
{
	char		buf[SIZE_BUFF+1]
		,		nbuf[SIZE_BUFF+1];
	ItemStruct	*code;
	char		*p;
	HakamaRecords	*recs = NULL;

ENTER_FUNC;
	CompileInit();
	strcpy(nbuf,name);
	if		(	(  name   ==  NULL  )
			||	(  *name  ==  0     ) ) {
		sprintf(buf,"%s\n",MSG_SCHEMA);
	} else {
		if		(  ( p = strchr(nbuf,'$') )  !=  NULL  ) {
			*p = 0;
			sprintf(buf,"%s%s:%s\n",MSG_SCHEMA,nbuf,p+1);
		} else {
			sprintf(buf,"%s%s\n",MSG_SCHEMA,name);
		}
	}
	dbgprintf("send [%s]",buf);
	if		(  !SendStringDelim(hakama->sd,buf)  )	return	NULL;
	RecvStringDelim(hakama->sd,SIZE_BUFF,buf);
	dbgprintf("recv [%s]", buf);
	code = CompileRecord(buf);
	if (code) {
	  recs = HakamaNewRecords();
	  ConvertRecord(recs,code);
	}
LEAVE_FUNC;
	return	(recs);
}

static	Bool
_ForDestroy(
	char			*key,
	HakamaKeyValue	*val)
{
	xfree(val->name);
	xfree(val->data);
	return	(TRUE);
}

extern	void
HakamaDestroyRecord(
	HakamaRecord	*rec)
{
ENTER_FUNC;
	if		(  rec  !=  NULL  ) {
		g_hash_table_foreach_remove(rec->item,(GHRFunc)_ForDestroy,NULL);
	}
	g_hash_table_destroy(rec->item);
	if		(  rec->vals  !=  NULL  ) {
		xfree(rec->vals);
	}
	xfree(rec);
LEAVE_FUNC;
}

extern	void
HakamaDestroyRecords(
	HakamaRecords	*recs)
{
	int		i;

ENTER_FUNC;
	if		(  recs  !=  NULL  ) {
		dbgprintf("rec->count = %d",recs->count);
		for	( i = 0 ; i < recs->count ; i ++ ) {
			if		(  recs->rec[i]  !=  NULL  ) {
				HakamaDestroyRecord(recs->rec[i]);
			}
		}
		xfree(recs);
	}
LEAVE_FUNC;
}

extern	void
HakamaMergeRecord(HakamaRecord *dest, HakamaRecord *src, int overwrite)
{
  int i;
  HakamaKeyValue *kv;
  
  for (i = 0; i < src->count; ++i) {
    kv = src->vals[i];
    if (overwrite) {
      	HakamaEnterValue(dest, kv->name, kv->data);
    } else {
      if (g_hash_table_lookup(dest->item, kv->name) == NULL) {
	HakamaEnterValue(dest, kv->name, kv->data);
      }
    }
  }
}

static int
HakamaIConv(
	    const char *fromcharset,
	    const char *tocharset,
	    char *outbuf,
	    int outbytes,
	    char *inbuf
)
{
  size_t inbytes = strlen(inbuf);
  size_t tmpbytes = inbytes * MB_LEN_MAX;
  char *tmpbuf_top = xmalloc(sizeof(char) * tmpbytes + 1);
  char *tmpbuf = tmpbuf_top;
  iconv_t cd = (iconv_t)-1;

  if (inbytes == 0) {
    if (outbytes > 0) {
      *outbuf = '\0';
    }
    goto end;
  }
  if (outbytes == 0) {
    goto end;
  }
  
  cd = iconv_open(tocharset, fromcharset);
  if (cd == (iconv_t)-1) {
    dbgprintf("iconv error: %s", strerror(errno));
    *outbuf = '\0';
    return -1;
  }

  iconv(cd, &inbuf, &inbytes, &tmpbuf, &tmpbytes);
  *tmpbuf = '\0';

  memset(outbuf, 0, outbytes);
  strncpy(outbuf, tmpbuf_top, outbytes - 1);

end:
  xfree(tmpbuf_top);
  if (cd != (iconv_t)-1) {
    iconv_close(cd);
  }

  return 0;
}

extern void
HakamaConvCharset(HakamaConvMode mode, char *outbuf, int outbytes, char *inbuf)
{
  if (strcmp(mode.from, mode.to) == 0) {
    memset(outbuf, 0, outbytes);
    strncpy(outbuf, inbuf, outbytes);
  } else {
    HakamaIConv(mode.from, mode.to, outbuf, outbytes, inbuf);
  }
}

extern void
HakamaConvCharsetRecord(HakamaConvMode convmode, HakamaRecord *rec)
{
  int i;
  HakamaKeyValue *kv;
  char *value;
  int nbytes;
  for (i = 0; i < rec->count; ++i) {
    kv = rec->vals[i];
    nbytes = strlen(kv->data) * MB_LEN_MAX + 1;
    value = (char *)xmalloc(nbytes);
    HakamaConvCharset(convmode, value, nbytes, kv->data);
    xfree(kv->data);
    kv->data = value;
  }
}

extern void
DestroyFields(Fields *fields)
{
  int i;
  if (!fields) {
    return;
  }

  if (fields->n >= 0) {
    for (i = 0; i < fields->n; ++i) {
      xfree(fields->name[i]);
    }
  }
  if (fields->name) {
    xfree(fields->name);
  }
  xfree(fields);
}
