/*
 * PANDA -- a simple transaction monitor
 * Copyright (C) 2000-2009 Ogochan & JMA (Japan Medical Association).
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/*
#define	CONSOLE_DEBUG
#define	DEBUG
#define	TRACE
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>
#include	"libs.h"
#include	"lex.h"
#include	"debug.h"

extern	void
UnGetChar(
	CURFILE	*in,
	int		c)
{
	in->pos --;
	in->back = c;
}

extern	int
GetChar(
	CURFILE	*in)
{
	int		c;

	if		(  in->back  >=  0  ) {
		c = in->back;
		in->back = -1;
	} else {
		if		(  in->pos  ==  in->size  ) {
			c = 0;
		} else {
			if		(  in->body  ==  NULL  ) {
				fprintf(stderr,"nulpo!\n");
			}
			if		(  ( c = in->body[in->pos] )  ==  0  ) {
				c = 0;
			}
		}
	}
	in->pos ++;
	return	(c);
}

extern	GHashTable	*
LexInit(
	TokenTable	*tokentable)
{
	int		i;
	GHashTable	*reserved;


ENTER_FUNC;
	reserved = NewNameiHash();
	for	( i = 0 ; tokentable[i].token  !=  0 ; i ++ ) {
		dbgprintf("[%s] = %d",tokentable[i].str,tokentable[i].token);
		g_hash_table_insert(reserved,StrDup(tokentable[i].str),
							(gpointer)tokentable[i].token);
	}
LEAVE_FUNC;
	return	(reserved);
}

static	int
CheckReserved(
	CURFILE	*in)
{
	gpointer	p;
	int		ret;
	char	ustr[SIZE_BUFF+1];
	char	*s;
	char	*str = in->Symbol;
	
	s = ustr;
	while	(  ( *s = toupper(*str) )  !=  0  ) {
		s ++;
		str ++;
	}
	if		(  ( p = g_hash_table_lookup(in->reserved,ustr) ) !=  NULL  ) {
		ret = (int)(long)p;
	} else {
		ret = T_SYMBOL;
	}
	dbgprintf("check reserved => [%s] -> %d", ustr, ret);
	
	return	(ret);
}

extern	int
Lex(
	CURFILE	*in,
	Bool	fName)
{
	int		c;
	char	*p;
	char	buff[SIZE_BUFF];

ENTER_FUNC;
  retry: 
	if		(  in->Symbol  !=  NULL  ) {
		xfree(in->Symbol);
		in->Symbol = NULL;
	}
	while	(	(  ( c = GetChar(in) )  !=  0  )
			&&	(  isspace(c)                  ) )	{
		if		(  c  ==  '\n'  ) {
			c = ' ';
			in->cLine ++;
		}
	}
	if		(  c  ==  '#'  ) {
		while	(	(  ( c = GetChar(in) )  !=  0    )
				&&	(  ( c = GetChar(in) )  !=  '\n' ) );
		in->cLine ++;
		goto	retry;
	}
	switch	(c) {
	  case	'"':
		p = buff;
		while	(  ( c = GetChar(in) )  !=  '"'  ) {
			if		(  c  ==  '\\'  ) {
				c = GetChar(in);
			}
			*p ++ = c;
		}
		*p = 0;
		in->Symbol = (char *)xmalloc(strlen(buff)+1);
		strcpy(in->Symbol,buff);
		in->Token = T_SCONST;
		break;
	  case	'\'':
		p = buff;
		while	(  ( c = GetChar(in) )  !=  '\''  ) {
			if		(  c  ==  '\\'  ) {
				c = GetChar(in);
			}
			*p ++ = c;
		}
		*p = 0;
		in->Symbol = (char *)xmalloc(strlen(buff)+1);
		strcpy(in->Symbol,buff);
		in->Token = T_SCONST;
		break;
	  case	0:
		in->Token = T_EOF;
		break;
	  default:
		p = buff;
		if		(	(  isalpha(c)  )
				||	(  c  ==  '_'  )
				||	(  c  ==  '$'  ) )	{
			do {
				*p ++ = c;
				c = GetChar(in);
			}	while	(	(  isalpha(c)  )
						||	(  isdigit(c)  )
						||	(  c  ==  '$'  )
						||	(  c  ==  '_'  ) );
			UnGetChar(in,c);
			*p = 0;
			in->Symbol = StrDup(buff);
			if		(  fName  ) {
				in->Token = T_SYMBOL;
			} else {
				in->Token = CheckReserved(in);
			}
		} else
		if		(  isdigit(c) )	{
			do {
				*p ++ = c;
				c = GetChar(in);
			}	while	(	(  isdigit(c)  )
						||	(  c  ==  '.'  ) );
			UnGetChar(in,c);
			*p = 0;
			in->Symbol = StrDup(buff);
			in->Int = atoi(in->Symbol);
			in->Token = T_ICONST;
		} else {
			switch	(c) {
			  case	EOF:
				in->Token = T_EOF;
				break;
			  default:
				in->Token = c;
				break;
			}
		}
		break;
	}
#ifdef	TRACE
	if		(  in->Token  >  0x7F  ) {
		dbgprintf("DB_Token = [%X][%s]\n",in->Token,in->Symbol);
	} else {
		dbgprintf("DB_Token = [%c]\n",in->Token);
	}
#endif
LEAVE_FUNC;
	return	(in->Token);
}

