/*
 * PANDA -- a simple transaction monitor
 * Copyright (C) 1995-2009  Peter Mattis, Spencer Kimball and Josh MacDonald
 *               2009  Ogochan & JMA (Japan Medical Association).
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/*
#define	CONSOLE_DEBUG
#define	TRACE
#define	DEBUG
*/

#include	<windows.h>
#include	<tchar.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	"libs.h"
#include	"debug.h"

#define HASH_TABLE_MIN_SIZE 11
#define HASH_TABLE_MAX_SIZE 13845163

typedef struct _GHashNode      GHashNode;

struct _GHashNode
{
	gpointer key;
	gpointer value;
	GHashNode *next;
};

struct _GHashTable
{
	gint size;
	gint nnodes;
	guint frozen;
	GHashNode **nodes;
	GHashFunc hash_func;
	GCompareFunc key_compare_func;
};

static void		g_hash_table_resize	 (GHashTable	*hash_table);
static GHashNode**	g_hash_table_lookup_node (GHashTable	*hash_table,
						  gconstpointer	 key);
static GHashNode*	g_hash_node_new		 (gpointer	 key,
						  gpointer	 value);
static void		g_hash_node_destroy	 (GHashNode	*hash_node);
static void		g_hash_nodes_destroy	 (GHashNode	*hash_node);

extern	void	*
_xmalloc(
	size_t	size,
	char	*fn,
	int		line,
	Bool	fClear)
{
	void	*ret;

	dbgprintf("xmalloc %s(%d),size = %d,",fn,line,size);
	if		(  ( ret = malloc(size) )  ==  NULL  )	{
		printf("no memory space!! %s(%d)\n",fn,line);
		exit(1);
	} else {
		if		(  fClear  ) {
			memclear(ret,size);
		}
	}
	return	(ret);
}

extern	void	*
_xrealloc(
	void	*p,
	size_t	size,
	char	*fn,
	int		line)
{
	void	*ret;

	dbgprintf("xrealloc %s(%d),size = %d,",fn,line,size);
	if		(  ( ret = realloc(p,size) )  ==  NULL  )	{
		printf("no memory space!! %s(%d)\n",fn,line);
		exit(1);
	}
	return	(ret);
}

extern	void
_xfree(
	void	*p,
	char	*fn,
	int		line)
{
	dbgprintf("xfree %s(%d)\n",fn,line);
	if		(  p  ==  NULL  )	return;
	free(p);
}

extern	char	*
StrDup(
	char	*s)
{
	char	*str;

	if		(  s  !=  NULL  ) {
		str = xmalloc(strlen(s)+1);
		strcpy(str,s);
	} else {
		str = NULL;
	}
	return	(str);
}

extern	char	*
IntStrDup(
	int		val)
{
	char	buff[20];

	sprintf(buff,"%d",val);
	return	(StrDup(buff));
}

static	long
HexToInt(
	char	*str,
	size_t	len)
{
	long	ret;

	ret = 0L;
	for	( ; len > 0 ; len -- )	{
		ret <<= 4;
		if		(  isdigit(*str)  )	{
			ret += ( *str - '0' );
		} else
		if		(  isxdigit(*str)  )	{
			ret += ( toupper(*str) - 'A' + 10 );
		}
		str ++;
	}
	return	(ret);
}

extern	char	*
StringChop(
	char	*str)
{
	char	*p;

	if		(  ( p = strrchr(str,'\r') )  !=  NULL  ) {
		*p = 0;
	}
	if		(  ( p = strrchr(str,'\n') )  !=  NULL  ) {
		*p = 0;
	}
	return	(str);
}

extern	size_t
DecodeStringURL(
	byte	*q,
	char	*p)
{
	byte	*qq = q;

	while	(	(  *p  !=  0    )
			&&	(  isspace(*p)  ) )	p ++;
	while	(  *p  !=  0  ) {
		if		(  *p  ==  '+'  ) {
			*q ++ = ' ';
		} else
		if		(  *p  ==  '%'  ) {
			*q ++ = (byte)HexToInt(p+1,2);
			p += 2;
		} else {
			*q ++ = *p;
		}
		p ++;
	}
	*q = 0;
	return	(q-qq);
}

extern	size_t
EncodeStringURL(
	char	*q,
	char	*p)
{
	char	*qq;

	qq = q;
	while	(  *p  !=  0  ) {
	  if		(  *(u_char*)p  ==  0x20  ) {
			*q ++ = '+';
		} else
		  if		(  isalnum(*(u_char*)p)  ) {
			*q ++ = *p;
		} else {
			*q ++ = '%';
			q += sprintf(q,"%02X",((int)*p)&0xFF);
		}
		p ++;
	}
	*q = 0;
	return	(q-qq);
}

static	GHashTable*
g_hash_table_new (
	GHashFunc		hash_func,
	GCompareFunc	key_compare_func)
{
	GHashTable *hash_table;
	guint i;
  
	hash_table = New0(GHashTable,1);
	hash_table->size = HASH_TABLE_MIN_SIZE;
	hash_table->nnodes = 0;
	hash_table->frozen = FALSE;
	hash_table->hash_func = hash_func;
	hash_table->key_compare_func = key_compare_func;
	hash_table->nodes = New2 (GHashNode*, hash_table->size);

	for (i = 0; i < hash_table->size; i++)
		hash_table->nodes[i] = NULL;
  
	return hash_table;
}

void
g_hash_table_destroy (GHashTable *hash_table)
{
	guint i;
  
	for (i = 0; i < hash_table->size; i++)
		g_hash_nodes_destroy (hash_table->nodes[i]);
  
	xfree (hash_table->nodes);
	xfree (hash_table);
}

static	inline	GHashNode**
g_hash_table_lookup_node (GHashTable	*hash_table,
						  gconstpointer	 key)
{
	GHashNode **node;
  
	node = &hash_table->nodes
		[(* hash_table->hash_func) (key) % hash_table->size];
  
	/* Hash table lookup needs to be fast.
	 *  We therefore remove the extra conditional of testing
	 *  whether to call the key_compare_func or not from
	 *  the inner loop.
	 */
	if (hash_table->key_compare_func) {
		while (*node && !(*hash_table->key_compare_func) ((*node)->key, key)) {
			dbgprintf("[%s]",(char *)(*node)->key);
			node = &(*node)->next;
		}
	} else {
		while (*node && (*node)->key != key)
			node = &(*node)->next;
	}

	return node;
}

extern	gpointer
g_hash_table_lookup (
	GHashTable	  *hash_table,
	gconstpointer key)
{
	GHashNode *node;
  
	node = *g_hash_table_lookup_node (hash_table, key);
  
	return node ? node->value : NULL;
}

extern	void
g_hash_table_insert(
	GHashTable *hash_table,
	gpointer	key,
	gpointer	value)
{
	GHashNode **node;
  
	node = g_hash_table_lookup_node (hash_table, key);
  
	if		(  *node  !=  NULL  )	{
		/* do not reset node->key in this place, keeping
		 * the old key might be intended.
		 * a g_hash_table_remove/g_hash_table_insert pair
		 * can be used otherwise.
		 *
		 * node->key = key; */
		(*node)->value = value;
    } else {
		*node = g_hash_node_new (key, value);
		hash_table->nnodes++;
		if (!hash_table->frozen)
			g_hash_table_resize (hash_table);
	}
}

extern	void
g_hash_table_remove (
	GHashTable		*hash_table,
	gconstpointer	key)
{
	GHashNode **node, *dest;
  
	node = g_hash_table_lookup_node (hash_table, key);

	if (*node) {
		dest = *node;
		(*node) = dest->next;
		g_hash_node_destroy (dest);
		hash_table->nnodes--;
  
		if (!hash_table->frozen)
			g_hash_table_resize (hash_table);
	}
}

extern	Bool
g_hash_table_lookup_extended (
	GHashTable		*hash_table,
	gconstpointer	lookup_key,
	gpointer		*orig_key,
	gpointer		*value)
{
	GHashNode *node;
  
	node = *g_hash_table_lookup_node (hash_table, lookup_key);
  
	if (node) {
		if (orig_key)
			*orig_key = node->key;
		if (value)
			*value = node->value;
		return TRUE;
	} else {
		return FALSE;
	}
}

extern	void
g_hash_table_freeze(
	GHashTable	*hash_table)
{
	hash_table->frozen++;
}

extern	void
g_hash_table_thaw (
	GHashTable	*hash_table)
{
	if (hash_table->frozen)
		if (!(--hash_table->frozen))
			g_hash_table_resize (hash_table);
}

extern	guint
g_hash_table_foreach_remove (
	GHashTable	*hash_table,
	GHRFunc		func,
	gpointer	user_data)
{
	GHashNode *node, *prev;
	guint i;
	guint deleted = 0;
  
	for (i = 0; i < hash_table->size; i++) {
	  restart:
      
		prev = NULL;
      
		for (node = hash_table->nodes[i]; node; prev = node, node = node->next) {
			if ((* func) (node->key, node->value, user_data)) {
				deleted += 1;
				hash_table->nnodes -= 1;
				if (prev) {
					prev->next = node->next;
					g_hash_node_destroy (node);
					node = prev;
				} else {
					hash_table->nodes[i] = node->next;
					g_hash_node_destroy (node);
					goto restart;
				}
			}
		}
	}
  
	if (!hash_table->frozen)
		g_hash_table_resize (hash_table);
  
	return deleted;
}

extern	void
g_hash_table_foreach (
	GHashTable	*hash_table,
	GHFunc		func,
	gpointer	user_data)
{
	GHashNode *node;
	gint i;
  
	for (i = 0; i < hash_table->size; i++)
		for (node = hash_table->nodes[i]; node; node = node->next)
			(* func) (node->key, node->value, user_data);
}

/* Returns the number of elements contained in the hash table. */
extern	guint
g_hash_table_size (
	GHashTable	*hash_table)
{
	return hash_table->nnodes;
}

static const guint g_primes[] = {
  11,
  19,
  37,
  73,
  109,
  163,
  251,
  367,
  557,
  823,
  1237,
  1861,
  2777,
  4177,
  6247,
  9371,
  14057,
  21089,
  31627,
  47431,
  71143,
  106721,
  160073,
  240101,
  360163,
  540217,
  810343,
  1215497,
  1823231,
  2734867,
  4102283,
  6153409,
  9230113,
  13845163,
};

static const guint g_nprimes = sizeof (g_primes) / sizeof (g_primes[0]);

static	guint
g_spaced_primes_closest (
	guint num)
{
	gint i;

	for (i = 0; i < g_nprimes; i++)
		if (g_primes[i] > num)
			return g_primes[i];

	return g_primes[g_nprimes - 1];
}

static void
g_hash_table_resize (
	GHashTable	*hash_table)
{
	GHashNode **new_nodes;
	GHashNode *node;
	GHashNode *next;
	float nodes_per_list;
	guint hash_val;
	gint new_size;
	gint i;

	nodes_per_list = (float) hash_table->nnodes / (float) hash_table->size;

	if ((nodes_per_list > 0.3 || hash_table->size <= HASH_TABLE_MIN_SIZE) &&
		(nodes_per_list < 3.0 || hash_table->size >= HASH_TABLE_MAX_SIZE))
		return;
  
	new_size = CLAMP(g_spaced_primes_closest (hash_table->nnodes),
					 HASH_TABLE_MIN_SIZE,
					 HASH_TABLE_MAX_SIZE);
	new_nodes = (GHashNode **)xmalloc(sizeof(GHashNode *) * new_size);
	for	( i = 0 ; i < new_size ; i ++ ) {
		new_nodes[i] = NULL;
	}
	for (i = 0; i < hash_table->size; i++)
		for (node = hash_table->nodes[i]; node; node = next) {
			next = node->next;
			hash_val = (* hash_table->hash_func) (node->key) % new_size;
			node->next = new_nodes[hash_val];
			new_nodes[hash_val] = node;
		}
  
	xfree(hash_table->nodes);

	hash_table->nodes = new_nodes;
	hash_table->size = new_size;

}

static	GHashNode*
g_hash_node_new (
	gpointer	key,
	gpointer	value)
{
	GHashNode *hash_node;

	hash_node = New(GHashNode);

	hash_node->key = key;
	hash_node->value = value;
	hash_node->next = NULL;
  
	return hash_node;
}

static	void
g_hash_node_destroy(
	GHashNode	*hash_node)
{
	xfree(hash_node);
}

static void
g_hash_nodes_destroy(
	GHashNode	*node)
{
	GHashNode	*next;

	if		(  node  !=  NULL  )	{
		while	(  node->next  !=  NULL  )	{
			next = node->next;
			xfree(node);
			node = next;
		}
	}
}

static	guint
NameHash(
	gconstpointer	key)
{
	char	*p;
	guint	ret;

	ret = 0;
	if		(  key  !=  NULL  ) {
		for	( p = (char *)key ; *p != 0 ; p ++ ) {
			ret = ( ret << 4 ) | *p;
		}
	}
	return	(ret);
}

static	gint
NameCompare(
	gconstpointer	s1,
	gconstpointer	s2)
{
	return	(!strcmp(s1,s2));
}

extern	GHashTable	*
NewNameHash(void)
{
	GHashTable	*ret;

	ret = g_hash_table_new((GHashFunc)NameHash,(GCompareFunc)NameCompare);

	return	(ret);
}

static	guint
NameiHash(
	gconstpointer	key)
{
	char	*p;
	guint	ret;

	ret = 0;
	if		(  key  !=  NULL  ) {
		for	( p = (char *)key ; *p != 0 ; p ++ ) {
			ret = ( ret << 1 ) | toupper(*p);
		}
	}
	return	(ret);
}

static	gint
NameiCompare(
	gconstpointer	s1,
	gconstpointer	s2)
{
dbgprintf("[%s] * [%s]",(char *)s1,(char *)s2);
	return	(!stricmp((char *)s1,(char *)s2));
}

extern	GHashTable	*
NewNameiHash(void)
{
	GHashTable	*ret;

	ret = g_hash_table_new((GHashFunc)NameiHash,(GCompareFunc)NameiCompare);

	return	(ret);
}
