/* GLIB - Library of useful routines for C programming
 * Copyright (C) 1995-1997  Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GLib Team and others 1997-1999.  See the AUTHORS
 * file for a list of people on the GLib Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GLib at ftp://ftp.gtk.org/pub/gtk/. 
 */

#ifndef __LIBS_H__
#define __LIBS_H__
#include	<windows.h>

#define	SIZE_BUFF		1024
#define	SIZE_LONGNAME	1024
#define	SIZE_LARGEBUFF	1024*1024

#define	Bool		int

#ifdef	FALSE
#undef	FALSE
#define	FALSE		0
#endif
#ifdef	TRUE
#undef	TRUE
#define	TRUE		(!FALSE)
#endif

typedef struct _GHashTable	GHashTable;
typedef void* gpointer;
typedef const void *gconstpointer;
typedef unsigned int	guint;
typedef int    gint;

#define	Bool	int

extern	void	*_xmalloc(size_t size, char *fn, int line, Bool fClear);
extern	void	*_xrealloc(void *p, size_t size, char *fn, int line);
extern	void	_xfree(void *p, char *fn, int line);
extern	char	*StringChop(char *str);
extern	size_t	DecodeStringURL(byte *q, char *p);
extern	size_t	EncodeStringURL(char *q, char *p);
extern	char	*StrDup(char *s);
extern	char	*IntStrDup(int val);

typedef gint		(*GCompareFunc)		(gconstpointer	a,
										 gconstpointer	b);
typedef guint		(*GHashFunc)		(gconstpointer	key);
typedef void		(*GFreeFunc)		(gpointer	data);
typedef void		(*GHFunc)		(gpointer	key,
						 gpointer	value,
						 gpointer	user_data);
typedef Bool	(*GHRFunc)		(gpointer	key,
						 gpointer	value,
						 gpointer	user_data);

extern	GHashTable	*NewNameHash(void);
extern	GHashTable	*NewNameiHash(void);

void	    g_hash_table_destroy	(GHashTable	*hash_table);
void	    g_hash_table_insert		(GHashTable	*hash_table,
					 gpointer	 key,
					 gpointer	 value);
void	    g_hash_table_remove		(GHashTable	*hash_table,
					 gconstpointer	 key);
gpointer    g_hash_table_lookup		(GHashTable	*hash_table,
					 gconstpointer	 key);
Bool    g_hash_table_lookup_extended(GHashTable	*hash_table,
					 gconstpointer	 lookup_key,
					 gpointer	*orig_key,
					 gpointer	*value);
void	    g_hash_table_freeze		(GHashTable	*hash_table);
void	    g_hash_table_thaw		(GHashTable	*hash_table);
void	    g_hash_table_foreach	(GHashTable	*hash_table,
					 GHFunc		 func,
					 gpointer	 user_data);
guint	    g_hash_table_foreach_remove	(GHashTable	*hash_table,
					 GHRFunc	 func,
					 gpointer	 user_data);
guint	    g_hash_table_size		(GHashTable	*hash_table);

#define New(type)				(type *)(xmalloc(sizeof(type)))
#define New2(type, count)		(type *)(xcalloc(sizeof(type) * (count)))
#define New0(type, count)		(type *)(xcalloc(sizeof(type) * (count)))
#undef	CLAMP
#define CLAMP(x, low, high)		(((x) > (high)) ? (high) : (((x) < (low)) ? (low) : (x)))

#define	memclear(b,s)			memset((b),(s),0)

#ifndef	xmalloc
#define	xmalloc(s)				_xmalloc((s),__FILE__,__LINE__,FALSE)
#endif
#ifndef	xrealloc
#define	xrealloc(p,s)			_xrealloc((p),(s),__FILE__,__LINE__)
#endif
#ifndef	xcalloc
#define	xcalloc(s)				_xmalloc((s),__FILE__,__LINE__,TRUE)
#endif
#ifndef	xfree
#define	xfree(p)				_xfree((p),__FILE__,__LINE__),(p) = NULL
#endif

#ifndef	strlcmp
#define	strlcmp(s1,s2)				strncmp(s1,s2,strlen(s2))
#endif



#endif
