# perl_sample.pl

# db server config: test1.db
# 
# test1   {
#        id      varchar(40);
#        class   varchar(20);
#        code    int;
# };
# 
# primary {
#        code;
# };
# path    id      {
#        DBSELECT        {
#                DECLARE test1_id_csr CURSOR FOR
#                SELECT  *
#                FROM test1
#                        WHERE
#                        (:code is NULL)
#                        OR code = :code
#                ;
#        };
# };
#

use strict;
use warnings;
use DBI;
use Data::Dumper;
use constant {
  TEST_CODE => 123456, #primary key
  TEST_ID => 'idtest',
  TEST_CLASS => 10
};
print "--- select example\n";
select_example();
print "--- edit example\n";
dui_example();

#print "-- select with driver option example\n";
#driver_select_example();

sub select_example
{
  # connect
  my $i = 0;
  my $dbh = DBI->connect('dbi:ODBC:Hakama DataSource');
  my $sth= $dbh->prepare('select code,id,class from test1$id') or die $dbh->errstr;
  
  $sth->execute or die $dbh->errstr;
  
  while (my $arrayref = $sth->fetchrow_arrayref) {
    print Dumper($arrayref);
  }
  $sth->finish;
  $dbh->disconnect;
}

sub driver_select_example
{
  # connect
  my $dbh = DBI->connect(
    'dbi:ODBC:Driver=Hakama ODBC Driver;'
    .'Server=localhost;Port=8013;'
    .'User=test;Password=test;'
    .'ODBCEncoding=CP932;DBSEncoding=EUC-JP;NonConvert=0') or die $!;
  my $sth = $dbh->prepare('select code,id,class from test1$id') or die $dbh->errstr;
  $sth->execute or die $dbh->errstr;
  
  while (my $arrayref = $sth->fetchrow_arrayref) {
    print Dumper $arrayref;
  }
  $sth->finish;
  $dbh->disconnect;
}

sub select_print_key
{
  my ($dbh, $key) = @_;
  my $stmt = $dbh->prepare('select code,id,class from test1$id where code=?;');
  $stmt->bind_param(1, $key);
  $stmt->execute;
  
  while (my $hashref = $stmt->fetchrow_hashref) {
    while (my($field, $value) = each %$hashref) {
      print "\t$field => $value\n";
    }
  }
  $stmt->finish;
}

sub dui_example
{
  # connect, AutoCommit => 0
  my $dbh = DBI->connect(
    'dbi:ODBC:Hakama DataSource',
    undef,
    undef,
    {
      AutoCommit => 0,
      RaiseError => 1, # die() if error
      PrintError => 0}
  ) or die $!;
  my $stmt;
  
  # insert
  print "insert..\n";
  $stmt = $dbh->prepare('insert into test1$id (code,id,class) values(?,NULL,NULL);');
  $stmt->bind_param(1, TEST_CODE);
  $stmt->execute;
  
  $dbh->commit;
  $stmt->finish;
  
  # select 1
  print "select 1..\n";
  select_print_key($dbh, TEST_CODE);
  
  # update
  print "update..\n";
  $stmt = $dbh->prepare('update test1$id set id=?, class=? where code=?;');
  $stmt->bind_param(1, TEST_ID);
  $stmt->bind_param(2, TEST_CLASS);
  $stmt->bind_param(3, TEST_CODE);
  $stmt->execute;
  
  $dbh->commit;
  $stmt->finish;
  
  # select 2
  print "select 2..\n";
  select_print_key($dbh, TEST_CODE);
  
  # delete
  print "delete ..\n";
  $stmt = $dbh->prepare('delete from test1$id1 where code=?;');
  $stmt->bind_param(1, TEST_CODE);
  $stmt->execute;
  
  $dbh->commit;
  $stmt->finish;
  
  # select 3
  print "select 3..\n";
  select_print_key($dbh, TEST_CODE);
  
  $dbh->disconnect;
}

