# ruby_sample.rb

# db server config: test1.db
# 
# test1   {
#        id      varchar(40);
#        class   varchar(20);
#        code    int;
# };
# 
# primary {
#        code;
# };
# path    id      {
#        DBSELECT        {
#                DECLARE test1_id_csr CURSOR FOR
#                SELECT  *
#                FROM test1
#                        WHERE
#                        (:code is NULL)
#                        OR code = :code
#                ;
#        };
# };
#

require 'rubygems'
require 'dbi'

$KCODE = 'Shift_JIS'

$TEST_CODE = 123456
$TEST_ID = 'idtest'
$TEST_CLASS = 10

def select_example
	begin
		dbh = DBI.connect('DBI:ODBC:Hakama DataSource')
		dbh.select_all('SELECT code,id,class FROM test1$id') do |row|
			p row
		end
	rescue DBI::DatabaseError => e
		puts "Error: #{e.err}: #{e.errstr}"
	ensure
		dbh.disconnect
	end
end

def driver_select_example
	begin
		dbh = DBI.connect(
			'dbi:ODBC:Driver=Hakama ODBC Driver;' <<
			'Server=localhost;Port=8013;' << 
			'User=test;Password=test;' << 
			'ODBCEncoding=CP932;DBSEncoding=EUC-JP;NonConvert=0'
		)
		dbh.select_all('SELECT code,id,class FROM test1$id') do |row|
			p row
		end
	rescue DBI::DatabaseError => e
		puts "Error: #{e.err}: #{e.errstr}"
	ensure
		dbh.disconnect
	end
end

def select_print_key(dbh, test_code)
	stmt = dbh.prepare('select code,id,class from test1$id where code=?;');
	stmt.execute(test_code);
	stmt.fetch do |row|
		row.each_with_name do |value, field|
	           printf "\t%s => %s\n", field, value.to_s
	       end
	end
	stmt.finish
end

def dui_example
	begin
		puts "insert .."
		dbh = DBI.connect('DBI:ODBC:Hakama DataSource')
		dbh['AutoCommit'] = false;
		
		stmt = dbh.prepare('insert into test1$id (code,id,class) values(?,?,?);')
		stmt.execute($TEST_CODE, nil, nil)
		dbh.commit
		stmt.finish
		
		puts "select 1 .."
		select_print_key(dbh, $TEST_CODE)
		
		puts "update .."
		stmt = dbh.prepare('update test1$id set id=?, class=? where code=?;')
		stmt.execute($TEST_ID, $TEST_CLASS, $TEST_CODE)
		dbh.commit
		stmt.finish
		
		puts "select 2 .."
		select_print_key(dbh, $TEST_CODE)
		
		puts "delete .."
		stmt = dbh.prepare('delete from test1$id1 where code=?;')
		stmt.execute($TEST_CODE)
		dbh.commit
		stmt.finish
		
		puts "select 3 .."
		select_print_key(dbh, $TEST_CODE)
	rescue DBI::DatabaseError => e
		puts "Error: #{e.err}: #{e.errstr}"
	ensure
		dbh.disconnect
	end
end

puts "--- select example"
select_example
puts "--- edit example"
dui_example
#puts "--- select with driver option example\n"
#driver_select_example

